import streamlit as st

from ovo import db, config
from ovo.app.components.navigation import project_round_selector, pool_selector_table
from ovo.app.utils.page_init import initialize_page
from ovo.app.components.create_new_pool import create_new_pool
from ovo.core.database import Design
from ovo.core.plugins import load_variable, plugin_design_views
from ovo.app.utils.cached_db import get_cached_pools_table, get_cached_rounds

initialize_page(page_title="Designs")

st.title("🐣 Designs")
project = st.session_state.project

st.write(f"Showing all design pools in project **{project.name}**")

rounds = get_cached_rounds(project_id=project.id)
rounds_by_id = {r.id: r for r in rounds}
if not rounds:
    st.write("*No designs available in this project. Upload a pool or submit a workflow in the left panel.*")
    if st.button(":material/upload: Upload designs", key="empty_upload"):
        create_new_pool()
    st.stop()


selected_round_ids, selected_design_ids = project_round_selector(rounds, allow_design_input=True)

if selected_design_ids is not None:
    selected_pool_ids = sorted(db.select_unique_values(Design, "pool_id", id__in=selected_design_ids))
else:
    st.subheader("Pools")

    if st.button(
        ":material/upload: Upload designs",
        disabled=config.props.read_only,
        help="Ovo is running in read-only mode, design upload is disabled" if config.props.read_only else None,
    ):
        create_new_pool()

    with st.spinner("Loading pools..."):
        pools_table = get_cached_pools_table(round_ids=selected_round_ids)

        if pools_table.empty:
            st.write("No pools created yet in this round")
            st.stop()

    selected_pool_ids = pool_selector_table(pools_table, st.session_state.project.id)

views = {
    "🔵 Explorer": "ovo.app.pages.designs.explorer:explorer_fragment",
    "🔎 ProteinQC": "ovo.app.pages.designs.proteinqc:proteinqc_fragment",
    "🎯 Interface analyzer": "ovo.app.pages.designs.interface:interface_fragment",
    "🔁 Refolding": "ovo.app.pages.designs.refolding:refolding_fragment",
    # "🧬 Structure prediction": "TODO",
    # "🧩 Plugins": "TODO",
}

# Update functions with
views.update(plugin_design_views)

if "design_view" not in st.session_state and "design_view" in st.query_params:
    st.session_state["design_view"] = st.query_params["design_view"]

view = st.segmented_control("Views", options=list(views), key="design_view")

if not view:
    if "design_view" in st.query_params:
        del st.query_params["design_view"]
    st.write(":material/arrow_upward: *Choose one of the view options to see your designs*")
    st.stop()

st.query_params["design_view"] = view

# Load function lazily to avoid importing all plugins when page is loaded
view_path = views[view]
assert isinstance(view_path, str), f"View function should be a string (module.submodule:func_name), got: {view_path}"
view_func = load_variable(view_path)

view_func(selected_pool_ids, design_ids=selected_design_ids)



import os
import traceback
from io import StringIO, BytesIO
from typing import List, Collection, Callable

import pandas as pd

from sqlalchemy.orm.attributes import flag_modified

from ovo import db, storage, get_scheduler, config
from ovo.core.auth import get_username
from ovo.core.database.descriptors_proteinqc import PROTEINQC_MAIN_DESCRIPTORS
from ovo.core.database.models_proteinqc import ProteinQCWorkflow
from ovo.core.database.models_refolding import RefoldingWorkflow, RefoldingSupportedDesignWorkflow
from ovo.core.database.descriptors import ALL_DESCRIPTORS_BY_KEY, ALL_DESCRIPTOR_KEYS_SET
from ovo.core.database.models import (
    DescriptorValue,
    Descriptor,
    Design,
    DescriptorJob,
    DescriptorWorkflow,
    DesignJob,
    DesignDescriptorWorkflow,
    Pool,
)
from ovo.core.logic.job_logic import update_job_status
from ovo.core.logic.proteinqc_logic import get_descriptor_cmap, get_descriptor_comment
from ovo.core.utils.export import write_sheet


def get_available_descriptors(design_ids: list[str]) -> dict[str, Descriptor]:
    """
    Return all descriptor keys found in DB for the given design ids.
    """
    design_ids = list(set(design_ids))
    if not design_ids:
        return {}
    descriptor_keys = db.select_unique_values(DescriptorValue, "descriptor_key", design_id__in=design_ids)
    return {
        descriptor_key: ALL_DESCRIPTORS_BY_KEY[descriptor_key]
        for descriptor_key in ALL_DESCRIPTORS_BY_KEY
        if descriptor_key in descriptor_keys
    }


def get_wide_descriptor_table(
    *,  # disallow positional arguments
    pool_ids: Collection[str] = None,
    design_ids: Collection[str] = None,
    descriptor_keys: Collection[str] = None,
    human_readable=True,
    nested=False,
    **filters,
) -> pd.DataFrame:
    """
    Return a wide descriptor table for the given design ids.
    The table will have design ids as index and human-readable descriptor names (or keys if human_readable=False) as columns.
    """
    if design_ids is None:
        if pool_ids is None:
            raise ValueError("Either pool_ids or design_ids must be provided")
        design_ids = db.select_values(Design, "id", pool_id__in=pool_ids, **filters)
    elif filters:
        raise ValueError(
            "Can only use use additional keyword filters when pool_ids are used, not when design_ids are used"
        )
    df = db.select_wide_descriptor_table(
        design_ids=design_ids,
        descriptor_keys=list(ALL_DESCRIPTORS_BY_KEY.keys()) if descriptor_keys is None else list(descriptor_keys),
    )
    if df.empty:
        return df

    # Reformat columns
    if human_readable:
        descriptors = [ALL_DESCRIPTORS_BY_KEY[descriptor_key] for descriptor_key in df.columns]
        if nested:
            df.columns = pd.MultiIndex.from_tuples((d.tool, d.name) for d in descriptors)
        else:
            df.columns = [d.name for d in descriptors]
    else:
        # avoid complicating the format combinations too much
        assert not nested, "nested=True is not supported when human_readable=False"

    # Get sequences from design spec
    # chain id -> design id -> sequence
    sequences_by_chain_and_design_id = {}
    for design_id, spec in db.select_dict(Design, "id", "spec", id__in=design_ids).items():
        for chain in spec.chains:
            chain_ids = ",".join(chain.chain_ids)
            if chain_ids not in sequences_by_chain_and_design_id:
                sequences_by_chain_and_design_id[chain_ids] = {}
            sequences_by_chain_and_design_id[chain_ids][design_id] = chain.sequence

    # Add sequence columns as (Sequence, A), (Sequence, B) etc
    for chain_ids, sequences_by_design_id in sequences_by_chain_and_design_id.items():
        if human_readable:
            seq_column = ("Sequence", chain_ids) if nested else f"Sequence {chain_ids}"
        else:
            assert not nested, "nested=True is not supported when human_readable=False"
            seq_column = f"sequence_{chain_ids}"
        df.insert(0, seq_column, pd.Series(sequences_by_design_id))

    return df


def submit_descriptor_workflow(workflow: DescriptorWorkflow, scheduler_key: str, project_id: str):
    if config.props.read_only:
        raise RuntimeError("Cannot submit design workflow: OVO server is in read-only mode")

    workflow.validate()

    username = get_username()

    scheduler = get_scheduler(scheduler_key)

    # Submit the workflow
    job_id = scheduler.submit(
        pipeline_name=workflow.get_pipeline_name(),
        params=workflow.prepare_params(workdir=scheduler.workdir),
    )

    # Create descriptor job
    descriptor_job = DescriptorJob(
        id=DescriptorJob.generate_id(),
        author=username,
        project_id=project_id,
        job_id=job_id,
        scheduler_key=scheduler_key,
        workflow=workflow,
    )

    # Save the descriptor job
    db.save(descriptor_job)
    return descriptor_job


def prepare_proteinqc_params(workflow: ProteinQCWorkflow, workdir: str) -> dict:
    storage_paths = []
    design_ids = []
    for design in db.select(Design, id__in=workflow.design_ids):
        if not design.structure_path:
            print(f"Design {design.id} has no pdb path. Skipping...")
            continue
        storage_paths.append(design.structure_path)
        design_ids.append(design.id)

    # Prepare a txt file with workflow input paths, each file renamed to design_id.pdb
    input_path = storage.prepare_workflow_inputs(storage_paths, workdir, names=design_ids)

    return {
        "input_pdb": input_path,
        "tools": ",".join(workflow.tools),
        "chains": ",".join(list(workflow.chains)),
        "batch_size": 50,
    }


def prepare_refolding_params(workflow: RefoldingWorkflow, workdir: str) -> dict:
    workflow.validate()

    designs = db.select(Design, id__in=workflow.design_ids)
    pools = db.select(Pool, id__in=set(d.pool_id for d in designs))
    design_workflows_by_pool_id = {
        p.id: db.get(DesignJob, id=p.design_job_id).workflow for p in pools if p.design_job_id
    }
    design_paths = {}
    for pool in pools:
        if not pool.design_job_id:
            # In case of custom designs, use design.structure_path as design input
            # TODO add warning if column seems to contain pLDDT - we don't want predicted structures as input!
            design_paths.update({d.id: d.structure_path for d in designs if d.pool_id == pool.id})
            continue
        design_workflow = design_workflows_by_pool_id[pool.id]
        if not isinstance(design_workflow, RefoldingSupportedDesignWorkflow):
            raise NotImplementedError(f"Workflow does not support refolding evaluation: {design_workflow.name}")
        pool_design_ids = [d.id for d in designs if d.pool_id == pool.id]
        design_paths.update(design_workflow.get_refolding_design_paths(pool_design_ids))

    if not design_paths:
        raise ValueError("No refolding structure paths available for the selected designs")

    # Prepare a txt file with workflow input paths, each file renamed to design_id.pdb
    input_designs_txt = storage.prepare_workflow_inputs(design_paths.values(), workdir, names=design_paths.keys())

    # Prepare native structure path (if specified)
    native_pdb_path = (
        storage.prepare_workflow_input(workflow.native_pdb_path, workdir) if workflow.native_pdb_path else None
    )

    # Prepare a single directory with reference files (requires filenames to be unique)
    return {
        "design_type": workflow.design_type,
        "input_designs": input_designs_txt,
        "native_pdb": native_pdb_path,
        "tests": ",".join(workflow.tests),
    }


def get_log(descriptor_job: DescriptorJob, tail: int = None) -> str:
    """Get the log of a descriptor job from the scheduler."""
    assert isinstance(descriptor_job, DescriptorJob), f"Expected DescriptorJob, got {type(descriptor_job).__name__}"
    scheduler = get_scheduler(descriptor_job.scheduler_key)
    log = scheduler.get_log(descriptor_job.job_id)
    if tail is not None:
        log_lines = log.splitlines()
        log = "\n".join(log_lines[-tail:])
    return log


def process_results(descriptor_job: DescriptorJob, callback: Callable = None, wait: bool = True):
    """Process results of a successful workflow - save DescriptorValues to database"""
    assert isinstance(descriptor_job, DescriptorJob), f"Expected DescriptorJob, got {type(descriptor_job).__name__}"
    if descriptor_job.processed:
        print("Job already processed")
        return
    scheduler = get_scheduler(descriptor_job.scheduler_key)
    if scheduler.get_result(descriptor_job.job_id) is None:
        # Job is still running
        if wait:
            print(f"Waiting for job {descriptor_job.job_id} to finish...")
            scheduler.wait(descriptor_job.job_id)
        else:
            raise ValueError(
                f"Job {descriptor_job.job_id} has not finished yet, try again later or use scheduler.wait(design_job.job_id)"
            )
    # Job should be successful or failed now, update job status
    update_job_status(descriptor_job)
    if descriptor_job.job_result is False:
        # Print job log and raise error
        print(scheduler.get_log(descriptor_job.job_id))
        raise ValueError(f"Job {descriptor_job.job_id} has failed")
    # Process and save descriptor values
    objects = descriptor_job.workflow.process_results(descriptor_job, callback=callback)
    assert isinstance(objects, list), f"Expected list from process_results(), got {type(objects).__name__}: {objects}"
    # Update processed flag
    descriptor_job.processed = True
    flag_modified(descriptor_job, "workflow")
    flag_modified(descriptor_job, "warnings")
    # Save all atomically in one commit
    db.save_all(objects + [descriptor_job])


def read_descriptor_file_values(
    descriptor_job: DescriptorJob,
    design_id_mapping: dict[str, str | tuple],
    filenames: dict[str, str] = None,
    descriptor_tables: dict[str, pd.DataFrame] = None,
) -> list[DescriptorValue]:
    """Process descriptor job, return list of DescriptorValues to be inserted into DB.

    :param descriptor_job: DescriptorJob object
    :param design_id_mapping: Mapping from design.id to table_id (basename of PDB file = id column in descriptor output file)
    :param filenames: Dict of "pipeline_name|tool_key" -> filename in output directory (without file extension - will look for .csv or .jsonl)
    :param descriptor_tables: Optional dictionary of pre-loaded descriptor tables (tool_key -> pd.DataFrame).
    """
    assert len(design_id_mapping) == len(set(design_id_mapping.values())), (
        f"Duplicate ids in design_id_mapping: {design_id_mapping}"
    )

    if descriptor_tables is None:
        descriptor_tables = {}

    if filenames:
        scheduler = get_scheduler(descriptor_job.scheduler_key)
        source_output_path = scheduler.get_output_dir(descriptor_job.job_id)

        batch_descriptors = {descriptor_key_prefix: [] for descriptor_key_prefix in filenames}
        contig_number = 1
        # Iterate over contigs until no more files are found
        while True:
            any_files_in_contig = False
            batch_number = 1
            # Iterate over batches until no more files are found
            while True:
                batch_name = f"contig{contig_number}_batch{batch_number}"
                any_files_in_batch = False
                for descriptor_key_prefix, filename in filenames.items():
                    df = None
                    jsonl_path = os.path.join(source_output_path, batch_name, f"{filename}.jsonl")
                    if storage.file_exists(jsonl_path):
                        try:
                            df = pd.read_json(StringIO(storage.read_file_str(jsonl_path)), lines=True)
                        except Exception as e:
                            raise ValueError(f"Failed to read {jsonl_path}: {e}")
                    csv_path = os.path.join(source_output_path, batch_name, f"{filename}.csv")
                    if storage.file_exists(csv_path):
                        df = pd.read_csv(StringIO(storage.read_file_str(csv_path)))
                    if df is not None:
                        if not df.empty:
                            id_column = find_id_column(df, descriptor_key_prefix)
                            batch_descriptors[descriptor_key_prefix].append(df.set_index(id_column))
                        any_files_in_batch = True
                        any_files_in_contig = True
                if not any_files_in_batch:
                    break
                batch_number += 1

            if not any_files_in_contig:
                break

            contig_number += 1

        if contig_number == 1 and batch_number == 1:
            raise ValueError(f"No descriptor files found in {source_output_path}")

        # Concatenate dataframes from all batches for each tool
        for key, dfs in batch_descriptors.items():
            if dfs:
                descriptor_tables[key] = pd.concat(dfs).sort_index()

    descriptor_values = []
    for design_id, table_ids in design_id_mapping.items():
        descriptor_values.extend(
            generate_descriptor_values_for_design(
                design_id=design_id,
                table_ids=table_ids,
                descriptor_job_id=descriptor_job.id,
                descriptor_tables=descriptor_tables,
                chains=descriptor_job.workflow.chains,
            )
        )

    return descriptor_values


def save_descriptor_job_for_design_job(
    design_job: DesignJob, project_id: str, chains: list[str], design_ids: list[str]
) -> DescriptorJob:
    descriptor_job = DescriptorJob(
        id=DescriptorJob.generate_id(),
        scheduler_key=design_job.scheduler_key,
        job_id=design_job.job_id,
        author=design_job.author,
        project_id=project_id,
        job_result=True,
        workflow=DesignDescriptorWorkflow(
            design_job_id=design_job.id,
            chains=chains,
            design_ids=design_ids,
        ),
    )
    # We need to save the descriptor job to obtain its autoincrement ID
    db.save(descriptor_job)
    return descriptor_job


def find_id_column(df: pd.DataFrame, df_name: str):
    for column in ["id", "ID", "Id"]:
        if column in df.columns:
            return column
    raise ValueError(f'Dataframe should contain an "ID" column, got: {df.columns} in {df_name}')


def generate_descriptor_values_for_design(
    design_id: str,
    table_ids: str | tuple,
    descriptor_job_id: str | None,
    descriptor_tables,
    chains: list[str],
) -> list[DescriptorValue]:
    """Generate DescriptorValue objects for a given design based on descriptor tables.

    Requirements (otherwise an error is raised, leading to interrupted processing of the job):
    - Each descriptor table must contain at least one recognized descriptor (defined by Descriptor objects)
    - One of the design's table_ids must be found in each descriptor table (each design should be present)
    - Descriptor table index must be unique

    :param design_id: ID of the design
    :param table_ids: ID(s) corresponding to the design in the dataframes (can be a single string or a tuple of strings)
    :param descriptor_job_id: ID of the DescriptorJob
    :param descriptor_tables: Dictionary of descriptor tables (tool_key -> pd.DataFrame, indexed by table_id)
    :param chains: List of chain IDs the descriptors apply to
    :return: List of DescriptorValue objects
    """
    if isinstance(table_ids, str):
        table_ids = (table_ids,)

    # Convert chains to comma-separated string
    if isinstance(chains, str):
        chains = list(chains.replace(",", ""))
    chains = ",".join(chains)

    descriptor_values = []
    for descriptor_key_prefix, descriptor_table in descriptor_tables.items():
        row = None
        for table_id in table_ids:
            if table_id in descriptor_table.index:
                row = descriptor_table.loc[table_id]
                assert not isinstance(row, pd.DataFrame), (
                    f"Found duplicate ID '{table_id}' in descriptor table '{descriptor_key_prefix}'"
                )
        if row is None:
            raise ValueError(
                f"ID {table_ids} ({design_id}) missing in {descriptor_key_prefix} descriptor table, found only: {descriptor_table.index.tolist()}"
            )
        recognized_fields = set()
        for field, value in row.items():
            if f"{descriptor_key_prefix}|{field}" not in ALL_DESCRIPTOR_KEYS_SET:
                continue
            recognized_fields.add(field)
            if pd.isna(value):
                continue
            descriptor_values.append(
                DescriptorValue(
                    design_id=design_id,
                    descriptor_key=f"{descriptor_key_prefix}|{field}",
                    descriptor_job_id=descriptor_job_id,
                    # TODO enable parsing row.chain(s) and store separate values for each chain or combination of chains
                    chains=chains,
                    value=str(value),
                )
            )
        if not recognized_fields:
            raise ValueError(f"No recognized descriptors found in {descriptor_key_prefix} table columns: {row.keys()}")
    return descriptor_values


def update_and_process_descriptors(descriptor_jobs: List[DescriptorJob], error_callback: Callable):
    """
    Update descriptor jobs and process finished descriptor jobs
    """
    for descriptor_job in descriptor_jobs:
        job_result = update_job_status(descriptor_job)

        if job_result is True and descriptor_job.workflow:
            workflow_name = descriptor_job.workflow.name or "workflow"
            try:
                process_results(descriptor_job, wait=False)
            except Exception as e:
                # Print exception and continue processing other jobs
                error_callback(
                    f"Unexpected error processing result of {workflow_name} (job {descriptor_job.job_id}): {e}"
                )
                traceback.print_exc()


def export_proteinqc_excel(design_ids: list[str], output_path: str = None):
    df = get_wide_descriptor_table(
        design_ids=design_ids,
        descriptor_keys=[d.key for d in PROTEINQC_MAIN_DESCRIPTORS],
        nested=False,
        human_readable=False,
    )
    # TODO add numbers of yellow and orange flags
    return export_design_descriptors_excel(df, output_path=output_path)


def export_design_descriptors_excel(df: pd.DataFrame, output_path=None) -> BytesIO | None:
    """Export Excel file with design descriptors for the given design ids.

    :param df: Dataframe with any values, descriptor columns should be descriptor keys
    :param output_path: If given, write the Excel file to this path. If None, return the bytes of the Excel file.
    """
    assert df.columns.nlevels == 1, "Expected nested=False dataframe with single-level columns"
    # Set index to include non-descriptor columns
    meta_cols = [col for col in df.columns if col not in ALL_DESCRIPTORS_BY_KEY]
    df = df.reset_index().set_index([df.index.name] + meta_cols)
    descriptors = [ALL_DESCRIPTORS_BY_KEY[c] for c in df.columns]

    # Rename columns to human-readable names
    if not df.empty:
        df.columns = pd.MultiIndex.from_tuples([(descriptor.tool, descriptor.name) for descriptor in descriptors])

    sheet_name = "Descriptors"
    buffer = BytesIO()
    writer = pd.ExcelWriter(buffer, engine="xlsxwriter")

    write_sheet(df, writer, sheet_name=sheet_name, index=True)

    workbook = writer.book
    sheet = writer.sheets[sheet_name]

    n_cols = sheet.dim_colmax + 1
    n_rows = sheet.dim_rowmax + 1

    row_offset = df.columns.nlevels
    column_offset = df.index.nlevels

    # Freeze panes
    sheet.freeze_panes(row_offset, column_offset)

    # Set column width
    index_width = 20
    sheet.set_column(0, 0, index_width)
    width = 10
    sheet.set_column(1, n_cols - 1, width)

    # Add filters
    sheet.autofilter(1, 0, n_rows - 1, n_cols - 1)

    # Apply background colors
    for i, (descriptor, col) in enumerate(zip(descriptors, df.columns), start=column_offset):
        cmap = get_descriptor_cmap(descriptor, min_val=df[col].dropna().min(), max_val=df[col].dropna().max())
        for row_idx, value in enumerate(df[col], start=row_offset):
            if pd.isna(value):
                sheet.write(row_idx, i, "")
            else:
                hex_color = cmap(value)
                cell_format = workbook.add_format({"bg_color": hex_color})
                sheet.write(row_idx, i, value, cell_format)

        # Write column description as a comment
        comment = get_descriptor_comment(descriptor)
        sheet.write_comment(1, i, comment, {"x_scale": 2})

    writer.close()
    if not output_path:
        buffer.seek(0)
        return buffer
    return None



import importlib
import importlib.metadata
import sys


def load_plugins():
    """Load all registered plugins"""
    # group label -> list of page dicts
    plugin_pages = {}
    # view name -> path to view (module.submodule:function_name)
    plugin_design_views = {}
    # list of paths to descriptor list (module.submodule:descriptors)
    plugin_descriptors = []
    # list of submodule names to load
    plugin_modules = []
    for entry_point in importlib.metadata.entry_points(group="ovo.plugins"):
        module_name = entry_point.value.split(":")[0]
        if entry_point.name == "plugin":
            print("Registering plugin {}".format(module_name), file=sys.stderr)
            plugin_dict = entry_point.load()
            pages = plugin_dict.get("pages", {})
            assert isinstance(pages, dict), "Plugin 'pages' should be a dict, got {} in {}".format(
                type(pages), module_name
            )
            for group_label, group_pages in pages.items():
                assert isinstance(group_pages, list), (
                    "Plugin 'pages' should be a dict of lists, got dict of {} in {}".format(
                        type(group_pages), module_name
                    )
                )
                plugin_pages[group_label] = plugin_pages.get(group_label, []) + group_pages
            for view_name, view_path in plugin_dict.get("design_views", {}).items():
                assert isinstance(view_path, str), (
                    "Expected view path to be a string (my_module.submodule:method_name), got {}".format(
                        type(view_path)
                    )
                )
                plugin_design_views[view_name] = view_path
            if descriptor_path := plugin_dict.get("descriptors"):
                assert isinstance(descriptor_path, str), (
                    "Expected descriptors path to be a string (my_module.submodule), got {}".format(
                        type(descriptor_path)
                    )
                )
                plugin_descriptors.append(descriptor_path)
            plugin_modules.extend(plugin_dict.get("modules", []))
    return plugin_pages, plugin_design_views, plugin_descriptors, plugin_modules


def load_variable(path):
    module_path, var_name = path.split(":")
    module = importlib.import_module(module_path)
    return getattr(module, var_name)


# Get list of plugin pages and import all plugins
plugin_pages, plugin_design_views, plugin_descriptors, plugin_modules = load_plugins()



import os
import shutil

from ovo import schedulers, NextflowScheduler, config

# Scheduler key to use for tests, must be set in environment
TEST_SCHEDULER_KEY = os.getenv("OVO_TEST_SCHEDULER", config.default_scheduler)
# Test output results will be symlinked to this directory
TEST_OUTPUT_DIR = os.getenv("OVO_TEST_OUTPUT_DIR", os.path.join(config.dir, "workdir", "test-results"))
# Name of testing project in the DB
TEST_PROJECT_NAME = "OVO Tests"


def get_test_scheduler() -> NextflowScheduler:
    if TEST_SCHEDULER_KEY not in schedulers:
        raise ValueError(
            f"Expected OVO_TEST_SCHEDULER to be one of {list(schedulers.keys())}, but got '{TEST_SCHEDULER_KEY}'"
        )
    return schedulers[TEST_SCHEDULER_KEY]


def prepare_test_output_dir(script_path: str) -> str:
    """Create test directory like test-results/workflows/rfdiffusion-end-to-end/test_scaffold_hairpin

    Removes existing directory if it exists.
    """
    path = script_path.removesuffix(".py")
    if "pipelines/" in path:
        path = path.split("pipelines/", 1)[-1]
        path = path.replace("/tests/", "/")
    else:
        path = os.path.basename(path)
    if test_name := os.getenv("PYTEST_CURRENT_TEST"):
        # append test function name to path
        test_name = test_name.split("::")[-1].split()[0]
        path = os.path.join(path, test_name)
    output_dir = os.path.join(TEST_OUTPUT_DIR, path)
    shutil.rmtree(output_dir, ignore_errors=True)
    return output_dir


def assert_similar_sequence(a: str, b: str, min_identity: float):
    """Assert that two sequences are similar enough."""
    if len(a) != len(b):
        raise AssertionError(f"Sequences have different lengths: {len(a)} != {len(b)}, {a} vs {b}")
    assert isinstance(min_identity, float) and 0 <= min_identity <= 1
    matches = sum(1 for x, y in zip(a, b) if x == y)
    identity = matches / len(a)
    if identity < min_identity:
        diff = f"{a}\n{b}\n" + "".join("^" if x != y else " " for x, y in zip(a, b))
        raise AssertionError(f"Sequences are not similar enough: {identity:.2f} < {min_identity:.2f}:\n{diff}")



import typer
import importlib
import os

app = typer.Typer(pretty_exceptions_enable=False)


@app.command()
def module(
    module_name: str = typer.Argument(None, help="Module name (will return path to 'ovo' module if not provided)"),
):
    """Get absolute path to installed python module in current environment"""
    if module_name is None:
        print(os.path.dirname(os.path.dirname(__file__)))
    else:
        module = importlib.import_module(module_name)
        print(os.path.dirname(module.__file__))



import os
from typing import Type, Sequence, Any
import pandas as pd

from sqlalchemy import create_engine, or_, func, case, text, distinct, inspect
from sqlalchemy.orm import Session, with_polymorphic
from sqlalchemy.orm.attributes import flag_modified

from ovo.core.database.base_db import T
from ovo.core.database.cache_clearing import CacheClearingEngine
from ovo.core.database.models import Base, DescriptorValue, Design
import sys


class SqlDBEngine(CacheClearingEngine):
    def __init__(self, db_url: str, verbose: bool = False, read_only: bool = False):
        if verbose:
            print("Connecting to", db_url, file=sys.stderr)
        self._db_url = db_url
        self._engine = create_engine(db_url, echo=verbose)
        self._read_only = read_only
        self._in_clause_items_limit = None
        if self._db_url.startswith("sqlite://"):
            # Sqlite IN clause is limited to 250k items (as of 2025)
            self._in_clause_items_limit = 200000

    def init(self):
        if self._db_url.startswith("sqlite:///"):
            db_path = self._db_url.removeprefix("sqlite:///")
            if not os.path.exists(db_path):
                os.makedirs(os.path.dirname(db_path), exist_ok=True)
        # Create the table(s)
        Base.metadata.create_all(self._engine)
        self.automigrate()

    def automigrate(self):
        with self._create_session() as session:
            inspector = inspect(self._engine)
            descriptor_job_columns = [c["name"] for c in inspector.get_columns("descriptor_job")]
            if "project_id" not in descriptor_job_columns and "round_id" in descriptor_job_columns:
                print("Applying automigration: adding project_id to descriptor_job", file=sys.stderr)
                session.execute(text("ALTER TABLE descriptor_job ADD COLUMN project_id VARCHAR"))
                # create index on project_id
                session.execute(text("CREATE INDEX ix_descriptor_job_project_id ON descriptor_job (project_id)"))
                # fill existing rows with project_id based on descriptor_job.round_id = round.id -> round.project_id
                session.execute(
                    text(
                        "UPDATE descriptor_job SET project_id = "
                        "(SELECT project_id FROM round WHERE round.id = descriptor_job.round_id)"
                    )
                )
                session.commit()
                session.execute(text("DROP INDEX IF EXISTS ix_descriptor_job_round_id"))
                session.execute(text("ALTER TABLE descriptor_job DROP COLUMN round_id"))
                session.commit()

    def _create_session(self) -> Session:
        return Session(bind=self._engine, expire_on_commit=False)

    def _check_updated(self, obj: Base):
        for field, value in obj.__dict__.items():
            if hasattr(value, "is_changed_hash"):
                if value.is_changed_hash():
                    flag_modified(obj, field)

    def check_read_only(self):
        if self._read_only:
            raise RuntimeError("Ovo is running in read-only mode, write operations are not allowed.")

    def save(self, obj: Base):
        self.check_read_only()
        super().save(obj)
        with self._create_session() as session:
            self._check_updated(obj)
            session.add(obj)
            session.commit()

    def save_all(self, objs: Sequence[Base]):
        self.check_read_only()
        super().save_all(objs)
        with self._create_session() as session:
            for obj in objs:
                self._check_updated(obj)
            session.add_all(objs)
            session.commit()

    def remove(self, model: Type[T], *id_args, **kwargs):
        self.check_read_only()
        super().remove(model, *id_args, **kwargs)
        assert id_args or kwargs, "At least one filter must be provided for remove operation"
        with self._create_session() as session:
            self._create_query(session, model=model, id_args=id_args, **kwargs).delete()
            session.commit()

    def save_value(self, model: Type[T], column: str, value, **kwargs):
        self.check_read_only()
        super().save_value(model, column, value, **kwargs)
        assert kwargs, "At least one filter must be provided for save_value operation"
        with self._create_session() as session:
            session.query(model).filter(*self._create_filters(model, kwargs)).update({column: value})
            session.commit()

    def select(self, model: Type[T], limit: int = None, order_by=None, **kwargs) -> Sequence[T]:
        with self._create_session() as session:
            return self._create_query(session, model=model, order_by=order_by, limit=limit, **kwargs).all()

    def count(self, model: Type[T], **kwargs) -> int:
        with self._create_session() as session:
            return self._create_query(session, model=model, **kwargs).count()

    def count_distinct(self, model: Type[T], field="id", group_by=None, **kwargs) -> int | dict[Any, int]:
        with self._create_session() as session:
            if group_by:
                if isinstance(group_by, str):
                    group_by = [group_by]
                group_by_fields = [getattr(model, column) for column in group_by]
                query = (
                    session.query(*(group_by_fields + [func.count(distinct(getattr(model, field)))]))
                    .filter(*self._create_filters(model, kwargs))
                    .group_by(*group_by_fields)
                )
                return {(row[0] if len(row) == 2 else tuple(row[:-1])): row[-1] for row in query.all()}
            else:
                return self._create_query(session, model=model, columns=[field], **kwargs).distinct().count()

    def _create_query(self, session, model: Type[T], columns=None, order_by=None, limit=None, id_args=None, **kwargs):
        if id_args:
            assert len(id_args) == 1, "Only one positional argument is allowed"
            kwargs["id"] = id_args[0]
        if columns:
            query = session.query(*[getattr(model, column) for column in columns])
        elif hasattr(model, "__mapper__") and model.__mapper__.polymorphic_on is not None:
            query = session.query(with_polymorphic(model, "*"))
        else:
            query = session.query(model)
        return (
            query.filter(*self._create_filters(model, kwargs))
            .order_by(*self._create_order_by(model, order_by))
            .limit(limit)
        )

    def _create_order_by(self, model: Type[T], order_by):
        if order_by is None:
            return []
        if isinstance(order_by, str):
            return [self._create_order_by_single(model, order_by)]
        return [self._create_order_by_single(model, k) for k in order_by]

    def _create_order_by_single(self, model: Type[T], order_by):
        return getattr(model, order_by) if not order_by.startswith("-") else getattr(model, order_by[1:]).desc()

    def _create_filters(self, model: Type[T], kwargs):
        filters = []
        for k, v in kwargs.items():
            # Handle or conditions
            if k == "_or":
                or_conditions = []
                for condition in v:
                    for or_k, or_v in condition.items():
                        or_conditions.append(getattr(model, or_k) == or_v)
                filters.append(or_(*or_conditions))
            # Handle in conditions
            elif k.endswith("__in"):
                descriptor_key = k[:-4]  # Remove the '__in' suffix
                filters.append(getattr(model, descriptor_key).in_(v))
            elif k.endswith("__ne"):
                descriptor_key = k[:-4]
                filters.append(getattr(model, descriptor_key) != v)
            else:
                filters.append(getattr(model, k) == v)
        return filters

    def get(self, model: Type[T], *id_args, **kwargs) -> T:
        with self._create_session() as session:
            return self._create_query(session, model=model, id_args=id_args, **kwargs).one()

    def get_value(self, model: Type[T], column: str, *id_args, raw=False, **kwargs):
        with self._create_session() as session:
            query = self._create_query(session, model=model, columns=[column], id_args=id_args, **kwargs)
            if raw:
                # Fetch raw DB column value, useful for fetching values that are now incompatible due to changes in models
                query = query.with_entities(text(column))
            return query.one()[0]

    def select_values(self, model: Type[T], column: str, order_by=None, **kwargs) -> list:
        with self._create_session() as session:
            return [
                row[0]
                for row in self._create_query(session, model=model, columns=[column], order_by=order_by, **kwargs).all()
            ]

    def select_dict(self, model: Type[T], key_column: str, value_column: str, order_by=None, **kwargs) -> dict:
        with self._create_session() as session:
            return {
                row[0]: row[1]
                for row in self._create_query(
                    session, model=model, columns=[key_column, value_column], order_by=order_by, **kwargs
                )
            }

    def select_unique_values(self, model: Type[T], column: str, **kwargs) -> set:
        with self._create_session() as session:
            return set(
                row[0]
                for row in self._create_query(
                    session,
                    model=model,
                    columns=[column],
                    **kwargs,
                ).all()
            )

    def select_dataframe(
        self, model: Type[T], index_col: str = "id", order_by=None, limit="unset", **kwargs
    ) -> pd.DataFrame:
        if limit == "unset":
            if not kwargs:
                raise ValueError(
                    "select_dataframe requires passing a filter, a limit, or explicit unlimited query limit=None to avoid fetching too many rows"
                )
            limit = None
        with self._create_session() as session:
            query = self._create_query(session, model=model, limit=limit, order_by=order_by, **kwargs)
            return pd.read_sql_query(query.statement, query.session.bind, index_col=index_col)

    def select_descriptor_values(self, descriptor_key: str, design_ids: list[str]) -> pd.Series:
        """Select values of a single descriptor for multiple designs."""
        # TODO this does not handle the case when a descriptor was computed multiple times for the same design
        #  same as in select_wide_descriptor_table below
        with self._create_session() as session:
            assert isinstance(descriptor_key, str), (
                f"Expected descriptor_key to be string, got {type(descriptor_key).__name__}"
            )
            values = {}
            design_ids = list(design_ids)
            batches = (
                [design_ids]
                if not self._in_clause_items_limit
                else [
                    design_ids[i : i + self._in_clause_items_limit]
                    for i in range(0, len(design_ids), self._in_clause_items_limit)
                ]
            )
            for batch in batches:
                query = (
                    session.query(DescriptorValue.design_id, DescriptorValue.value)
                    .filter(DescriptorValue.descriptor_key == descriptor_key)
                    .filter(*self._create_filters(DescriptorValue, dict(design_id__in=batch)))
                )
                values.update({design_id: value for design_id, value in query})
            series = pd.Series(values)
            try:
                # Try converting whole array to numbers.
                # If any value is not a number, keep it as array of strings.
                # This mimics original behavior of to_numeric(errors="ignore").
                series = pd.to_numeric(series)
            except ValueError:
                pass
            # Reindex to keep the original order, adding NaN for missing designs
            return series.reindex(design_ids)

    def select_design_descriptors(self, design_id: str, descriptor_keys: list[str]) -> pd.Series:
        """Select values of multiple descriptors for a single design."""
        # TODO this does not handle the case when a descriptor was computed multiple times for the same design
        #  same as in select_wide_descriptor_table below
        with self._create_session() as session:
            for key in descriptor_keys:
                assert isinstance(key, str), f"Expected collection of descriptor key strings, got {type(key).__name__}"
            query = (
                session.query(DescriptorValue.descriptor_key, DescriptorValue.value)
                .filter(DescriptorValue.design_id == design_id)
                .filter(*self._create_filters(DescriptorValue, {"descriptor_key__in": descriptor_keys}))
            )
            result = query.all()
            series = pd.Series({descriptor_key: value for descriptor_key, value in result})
            try:
                # Try converting whole array to numbers.
                # If any value is not a number, keep it as array of strings.
                # This mimics original behavior of to_numeric(errors="ignore").
                series = pd.to_numeric(series)
            except ValueError:
                pass
            return series.reindex(descriptor_keys)

    def select_wide_descriptor_table(self, design_ids: list[str], descriptor_keys: list[str], **kwargs) -> pd.DataFrame:
        # TODO this does not handle the case when a descriptor was computed multiple times for the same design
        #  this can happen when we implement multiple descriptor jobs with different settings.
        #  To solve this, some descriptor job key could be incorporated in the column name or used as a filter.
        with self._create_session() as session:
            for key in descriptor_keys:
                assert isinstance(key, str), f"Expected collection of descriptor keys strings, got {type(key).__name__}"
            design_ids = list(design_ids)
            batches = (
                [design_ids]
                if not self._in_clause_items_limit
                else [
                    design_ids[i : i + self._in_clause_items_limit]
                    for i in range(0, len(design_ids), self._in_clause_items_limit)
                ]
            )
            results = []
            for batch in batches:
                cases = [
                    func.max(case((DescriptorValue.descriptor_key == key, DescriptorValue.value), else_=None)).label(
                        key
                    )
                    for key in descriptor_keys
                ]

                query = (
                    session.query(DescriptorValue.design_id, *cases)
                    .filter(*self._create_filters(DescriptorValue, dict(design_id__in=batch)))
                    .filter(*self._create_filters(DescriptorValue, kwargs))
                    .group_by(DescriptorValue.design_id)
                )
                results += query.all()
            session.close()
            df = pd.DataFrame(results)
            if not df.empty:
                df = df.set_index("design_id")
            # remove columns with no values
            df = df.dropna(axis=1, how="all")
            for col in df.columns:
                try:
                    # Try converting whole array to numbers.
                    # If any value is not a number, keep it as array of strings.
                    # This mimics original behavior of to_numeric(errors="ignore").
                    df[col] = pd.to_numeric(df[col])
                except (ValueError, TypeError):
                    pass
            return df.reindex(design_ids)

    def get_design_accepted_values(self, design_ids: list[str]):
        with self._create_session() as session:
            query = self._create_query(session, model=Design, columns=["id", "accepted"], id__in=design_ids)
            result = query.all()  # Returns list of tuples (id, accepted)

        series = pd.Series({design_id: value for design_id, value in result})
        return series.reindex(design_ids)



import os
import uuid
from abc import ABC, abstractmethod
from copy import copy
from dataclasses import dataclass, field, asdict
from datetime import datetime
from datetime import timezone
from typing import Callable, List, Literal, Union

import pandas as pd
from sqlalchemy import String, Boolean, DateTime, func, Integer, UniqueConstraint, JSON
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import MappedAsDataclass
from sqlalchemy.orm import mapped_column

from ovo.core.database.encoder import DataclassType
from ovo.core.database.db_proxy import DBProxy
from ovo.core.scheduler.base_scheduler import Scheduler
from ovo.core.utils.formatting import generate_id
from ovo.core.utils.pdb import (
    get_sequences_from_pdb_str,
    ChainNotFoundError,
    get_standardized_remarks_from_pdb_str,
)
import json


class Base(MappedAsDataclass, DeclarativeBase, metaclass=DBProxy):
    """subclasses will be converted to dataclasses"""


class JobMixin(MappedAsDataclass):
    scheduler_key: Mapped[str] = mapped_column(String, default=None, nullable=False, index=True)
    job_id: Mapped[str] = mapped_column(String, default=None, nullable=False, index=True)
    job_started_date_utc: Mapped[datetime] = mapped_column(DateTime, default=None, nullable=True)
    job_finished_date_utc: Mapped[datetime] = mapped_column(DateTime, default=None, nullable=True)
    # None = in progress, True = success, False = failed
    job_result: Mapped[bool | None] = mapped_column(Boolean, default=None, nullable=True, index=True)
    warnings: Mapped[list[str]] = mapped_column(JSON, default_factory=list, nullable=False)


class MetadataMixin(MappedAsDataclass):
    author: Mapped[str] = mapped_column(String, default=None, nullable=False, index=True)
    created_date_utc: Mapped[datetime] = mapped_column(
        DateTime,
        nullable=False,
        index=True,
        default_factory=lambda: datetime.now(timezone.utc),
        server_default=func.datetime("now"),
    )


class UserSettings(Base):
    __tablename__ = "user_setting"

    username: Mapped[str] = mapped_column(String, primary_key=True)
    last_project_id: Mapped[str] = mapped_column(String, default=None, nullable=True)


class Project(Base, MetadataMixin):
    __tablename__ = "project"
    __table_args__ = (UniqueConstraint("name", "author", name="uq_project_name_author"),)

    id: Mapped[str] = mapped_column(String, primary_key=True, default_factory=lambda: str(uuid.uuid4()))
    # TODO should project names be unique? Or conditionally - if they are public?
    name: Mapped[str] = mapped_column(String, default=None, nullable=False)
    public: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False, index=True)


class Round(Base, MetadataMixin):
    __tablename__ = "round"

    id: Mapped[str] = mapped_column(String, primary_key=True, default_factory=lambda: str(uuid.uuid4()))
    project_id: Mapped[str] = mapped_column(String, default=None, nullable=False, index=True)
    name: Mapped[str] = mapped_column(String, default=None, nullable=False)
    description: Mapped[str] = mapped_column(String, default=None, nullable=True)


MODEL_WEIGHTS_SCAFFOLD = ["Base", "ActiveSite"]
MODEL_WEIGHTS_BINDER = ["Complex_base", "Complex_beta"]


@dataclass
class WorkflowParams:
    def validate(self):
        pass

    @classmethod
    def from_dict(cls, data):
        return cls(**data)

    def to_dict(self, human_readable=False):
        data = asdict(self)
        if human_readable:
            data = {k: v for k, v in data.items() if self.__dataclass_fields__[k].metadata.get("show_to_user", True)}
        return data


@dataclass
class Threshold:
    # min value
    min_value: float | None = None
    # max value
    max_value: float | None = None
    # enabled flag
    enabled: bool = True

    def copy(self, **kwargs) -> "Threshold":
        copied = copy(self)
        for key, value in kwargs.items():
            setattr(copied, key, value)
        return copied

    def __eq__(self, other):
        # NOTE: Not checking type because it's broken by live-reload, check attributes instead
        if not hasattr(other, "min_value") or not hasattr(other, "max_value") or not hasattr(other, "enabled"):
            return False
        return self.min_value == other.min_value and self.max_value == other.max_value and self.enabled == other.enabled

    def format(self, descriptor_name: str = "") -> str | None:
        if self.min_value is None and self.max_value is not None:
            return f"{descriptor_name} ≤ {self.max_value:.3f}".strip()
        elif self.max_value is None and self.min_value is not None:
            return f"{descriptor_name} ≥ {self.min_value:.3f}".strip()
        elif self.min_value is not None and self.max_value is not None:
            if descriptor_name:
                return f"{descriptor_name} within {self.min_value:.3f} — {self.max_value:.3f}"
            else:
                return f"{self.min_value:.3f} — {self.max_value:.3f}"
        return None

    def get_bounds(
        self, descriptor: "NumericDescriptor", descriptor_values: pd.Series
    ) -> tuple[float | int, float | int] | None:
        """Get min, max bounds for the threshold for a specific set of designs, based on the descriptor values

        Can be used to plot the threshold on a histogram or scatter plot.

        :param descriptor: Descriptor object, used to get the min and max values if not set in the threshold
        :param descriptor_values: Series of descriptor values for the designs, used to determine the bounds
        :return: tuple of (min, max) bounds for the threshold, or None if no bounds are set
        """
        if self.min_value is None and self.max_value is None:
            return None
        # we will widen the range for integer values to make it clear that the value is included (we use >= and <=)
        is_int = all(int(v) == v for v in descriptor_values if not pd.isna(v))
        min_allowed, max_allowed = self.min_value, self.max_value
        # delta = 1.5 times the whole X range of the plot
        delta = (descriptor_values.max() - descriptor_values.min()) * 1.5 if not descriptor_values.empty else 1
        if min_allowed is None:
            # (-inf to threshold.max_value)
            if descriptor.min_value is not None:
                # if the descriptor has a defined min value we can start from there
                min_allowed = descriptor.min_value
            else:
                # otherwise we start at the max minus the width of the plot (we double the range)
                min_allowed = max_allowed - delta
            if not descriptor_values.empty and descriptor_values.min() < min_allowed:
                # expand the range further in case any actual values are even lower
                min_allowed = descriptor_values.min()
        elif max_allowed is None:
            # (threshold.min_value to +inf)
            if descriptor.max_value is not None:
                max_allowed = descriptor.max_value
            else:
                max_allowed = min_allowed + delta
            if not descriptor_values.empty and descriptor_values.max() > max_allowed:
                max_allowed = descriptor_values.max()
        else:
            # threshold.min_value to threshold.max_value, we are all set
            pass

        if is_int:
            min_allowed -= 0.5
            max_allowed += 0.5

        return min_allowed, max_allowed


class WorkflowTypes:
    # workflow name -> class
    _registry = {}

    @classmethod
    def register(cls, workflow_name):
        """Decorator to register a workflow class with a given name"""

        def decorator(registered_class):
            cls._registry[workflow_name] = registered_class
            registered_class.name = workflow_name
            return registered_class

        return decorator

    @classmethod
    def exists(cls, workflow_name):
        """Check if a workflow type with the given name exists"""
        return workflow_name in cls._registry

    @classmethod
    def get(cls, workflow_name):
        """Get workflow class by name"""
        if workflow_name not in cls._registry:
            raise ValueError(
                f"Workflow type '{workflow_name}' is not registered, available types: {list(cls._registry.keys())}"
            )
        return cls._registry[workflow_name]

    @classmethod
    def get_subclass_names(cls, workflow_name) -> list[str]:
        """Get list of registered workflow names that are subclasses of the given workflow name (including itself)"""
        BaseWorkflowType = WorkflowTypes.get(workflow_name)
        workflow_names = [workflow_name]
        for n, WorkflowType in WorkflowTypes._registry.items():
            if n == workflow_name:
                # workflow_name is already included at the start of the list
                continue
            if WorkflowType.is_subclass(BaseWorkflowType):
                workflow_names.append(n)
        return workflow_names


@dataclass
class Workflow:
    # name is filled in by @WorkflowTypes.register decorator
    name: str = field(init=False)

    def is_instance(self, cls):
        """Safe isinstance that avoids error on objects loaded from Streamlit session state after live reload

        This solves the problem that sometimes is_instance returns false in our app
        even though the object actually is of that class.

        See documentation: https://docs.streamlit.io/develop/concepts/design/custom-classes
        And issue: https://github.com/streamlit/streamlit/issues/8180
        """
        return self.__class__.is_subclass(cls)

    @classmethod
    def is_subclass(cls, other_cls):
        """Safe issubclass that avoids error on objects loaded from Streamlit session state after live reload"""
        other_cls_name = f"{other_cls.__module__}.{other_cls.__name__}"
        this_class_and_superclass_names = [f"{c.__module__}.{c.__name__}" for c in cls.__mro__]
        return other_cls_name in this_class_and_superclass_names

    @classmethod
    def from_dict(cls, data):
        """Convert dict to Workflow object - NOTE: This method currently cannot be extended, the base class method will always be called! See db.encoder module"""

        data = copy(data)
        workflow_name = data["name"]
        if not WorkflowTypes.exists(workflow_name):
            return UnknownWorkflow(
                data=data,
                error=f"Workflow is not registered: '{workflow_name}'",
            )

        try:
            subclass = WorkflowTypes.get(workflow_name)
            for field_name, empty_params in subclass().get_param_fields().items():
                if data.get(field_name):
                    # instantiate params dataclasses from dicts
                    data[field_name] = type(empty_params).from_dict(data[field_name])

            # convert all thresholds to Threshold objects
            if "acceptance_thresholds" in data:
                for key, value in data["acceptance_thresholds"].items():
                    data["acceptance_thresholds"][key] = Threshold(**value)

            return subclass(**{k: v for k, v in data.items() if k not in ("name",)})
        except Exception as e:
            if os.getenv("OVO_STRICT_JSON", "").lower() in ("1", "true", "yes"):
                raise e
            return UnknownWorkflow(
                data=data,
                error=f"Unexpected error when loading workflow from DB, this might be due to changes to the schema. Migration steps might need to be applied. Error: {e}",
            )

    def get_param_fields(self, human_readable=False) -> dict[str, WorkflowParams]:
        """Get all ..._param fields, verify they are instances of WorkflowParams

        with human_readble=False, returns dict: field_name -> params instance
        with human_readble=True, returns dict: Tool name -> params instance
        """
        fields = {}
        for field_name, field_def in self.__dataclass_fields__.items():
            value = getattr(self, field_name)
            if tool_name := field_def.metadata.get("tool_name"):
                assert value is not None, (
                    f"Params should be initialized with default factory, "
                    f"found None in {type(self).__name__}.{field_name}"
                )

                # NOTE: if this fails, it might also be due to live-reload,
                # we might need to adapt the logic to handle that
                assert isinstance(value, WorkflowParams), "Params should be instance of WorkflowParams"
                assert "__dataclass_fields__" in type(value).__dict__, (
                    f"WorkflowParams subclass should be decorated with @dataclass: {type(value).__name__}"
                )
                key = tool_name if human_readable else field_name
                fields[key] = value
            elif isinstance(value, WorkflowParams):
                raise ValueError(
                    f'Workflow param fields should be declared with field(metadata=dict(tool_name="...")) '
                    f"in {type(self).__name__}.{field_name}"
                )

        return fields

    def get_table_row(self, **kwargs) -> pd.Series:
        """Get all values of all param fields, skip fields with metadata.show_to_user=False, return pd.Series"""
        if param_fields := self.get_param_fields(human_readable=True):
            param_columns = {
                (tool_name, subfield_name): value
                for tool_name, params in param_fields.items()
                for subfield_name, value in params.to_dict(human_readable=True).items()
            }
        elif hasattr(self, "params") and isinstance(self.params, dict):
            field_def = self.__dataclass_fields__.get("params")
            if field_def and field_def.metadata and "tool_name" in field_def.metadata:
                tool_name = field_def.metadata["tool_name"]
            else:
                tool_name = "Params"
            param_columns = {(tool_name, k): v for k, v in self.params.items()}
        else:
            param_columns = {}
        return pd.Series(
            {
                ("Workflow", "type"): self.name,
                **param_columns,
            },
            **kwargs,
        )

    def validate(self):
        for field_name, params in self.get_param_fields().items():
            try:
                params.validate()
            except Exception as e:
                raise ValueError(f"Validation failed for workflow.{field_name} ({type(params).__name__}: {e}") from e

    def get_time_estimate(self, scheduler: Scheduler) -> str:
        return "Time estimate: No time estimate available for this workflow"

    @abstractmethod
    def get_pipeline_name(self) -> str:
        """Get name (for example ovo.rfdiffusion-end-to-end) or github URL of pipeline to be submitted for this workflow"""
        raise NotImplementedError()

    @abstractmethod
    def prepare_params(self, workdir: str) -> dict:
        """Submit the workflow to the scheduler and return the job id"""
        raise NotImplementedError()

    @abstractmethod
    def process_results(self, job: Union["DesignJob", "DescriptorJob"], callback: Callable = None) -> list[Base]:
        """Process results of a successful workflow - download files from workdir,
        create and return objects to be saved such as Designs and DescriptorValues"""
        raise NotImplementedError()

    @classmethod
    def get_download_fields(cls):
        return {
            "pdb files": (Design, "structure_path"),
        }


@dataclass
class DescriptorWorkflow(Workflow, ABC):
    chains: List[str] = field(default_factory=list)
    design_ids: List[str] = field(default_factory=list)

    def validate(self):
        super().validate()
        if not self.chains:
            raise ValueError("No chains provided")
        if not isinstance(self.chains, list):
            raise ValueError(
                f"DescriptorWorkflow.chains must be a list, got {self.chains} ({type(self.chains).__name__})"
            )
        if not self.design_ids:
            raise ValueError("No design IDs provided")


@WorkflowTypes.register("Design Descriptor Workflow")
@dataclass
class DesignDescriptorWorkflow(DescriptorWorkflow):
    """Placeholder DescriptorWorkflow used when the descriptors are generated by a design workflow"""

    design_job_id: str = field(default=None)


@dataclass
class DesignWorkflow(Workflow):
    # acceptance threshold values (descriptor key -> interval (min, max, enabled))
    acceptance_thresholds: dict[str, Threshold] = field(default_factory=dict)

    def get_table_row(self, **kwargs) -> pd.Series:
        """Get all values of all param fields, skip fields with metadata.show_to_user=False, return pd.Series"""
        row = super().get_table_row(**kwargs)
        from ovo.core.database.descriptors import ALL_DESCRIPTORS_BY_KEY

        return pd.concat(
            [
                row,
                pd.Series(
                    {
                        ("Thresholds", ALL_DESCRIPTORS_BY_KEY[key].name): t.format()
                        for key, t in self.acceptance_thresholds.items()
                        if t.enabled
                    },
                    **kwargs,
                ),
            ]
        )

    def validate(self):
        super().validate()
        from ovo.core.database.descriptors import ALL_DESCRIPTORS_BY_KEY

        for key, threshold in self.acceptance_thresholds.items():
            if not isinstance(threshold, Threshold):
                raise ValueError(
                    f"Acceptance threshold for {key} should be an instance of Threshold, got {threshold} ({type(threshold).__name__})"
                )
            if not isinstance(key, str):
                raise ValueError(f"Acceptance threshold key should be a string, got {key} ({type(key).__name__})")
            if key not in ALL_DESCRIPTORS_BY_KEY:
                raise ValueError(f"Unknown descriptor key '{key}' in acceptance thresholds")

    @classmethod
    def visualize_multiple_designs_structures(cls, design_ids: list[str]):
        """Visualize multiple design structures in Streamlit"""
        raise NotImplementedError()

    @classmethod
    def visualize_single_design_structures(cls, design_id: str):
        """Visualize single design structures in Streamlit"""
        from ovo.app.components.workflow_visualization_components import visualize_design_structure

        visualize_design_structure(design_id)

    @classmethod
    def visualize_single_design_sequences(self, design_id: str):
        from ovo.app.components.workflow_visualization_components import visualize_design_sequence

        visualize_design_sequence(design_id)


@WorkflowTypes.register("Unknown Workflow")
class UnknownWorkflow(DesignWorkflow, DescriptorWorkflow):
    """Unknown workflow type, used when the workflow type is not recognized or not registered

    This can happen when a plugin is uninstalled.
    """

    # Avoid saving this class to DB, handled by DataclassEncoder
    __do_not_serialize__ = True
    # Raw data dict as stored in the workflow column
    data: dict = None
    # Explanation for why this workflow couldn't be loaded
    error: str = None

    def __init__(self, data: dict, error: str):
        self.data = data
        self.error = error
        # inherit all fields from DesignWorkflow and DescriptorWorkflow, combined inheritance doesn't seem to do the trick
        self.acceptance_thresholds = {}
        self.chains = []
        self.design_ids = []


class DesignJob(Base, MetadataMixin, JobMixin):
    __tablename__ = "design_job"

    id: Mapped[str] = mapped_column(String, primary_key=True, default_factory=lambda: str(uuid.uuid4()))
    workflow: Mapped[DesignWorkflow] = mapped_column(DataclassType(Workflow), default=None, nullable=False)


class Pool(Base, MetadataMixin):
    __tablename__ = "pool"
    __table_args__ = (UniqueConstraint("name", "round_id", name="uq_pool_name_round"),)

    id: Mapped[str] = mapped_column(String, primary_key=True, default=None)
    round_id: Mapped[str] = mapped_column(String, default=None, nullable=False, index=True)
    design_job_id: Mapped[str] = mapped_column(String, default=None, nullable=True, index=True)
    name: Mapped[str] = mapped_column(String, default=None, nullable=False)
    description: Mapped[str] = mapped_column(String, default=None, nullable=True)
    # flag used to easily recognize a pool that has just finished but has not been processed yet
    # (job_result = True and processed = False)
    processed: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)

    @classmethod
    def generate_id(cls):
        return generate_id(previous_ids=Pool.select_unique_values("id"))


DesignChainType = Literal["protein", "dna", "rna", "ligand"]


x = {
    "id": "asdads",
    "spec": {
        "chains": [
            {"chain_ids": ["H"], "sequence": "QVQLV..."},
            {"chain_ids": ["L"], "sequence": "EIVL..."},
        ]
    },
    "structure_path": "asdasd/asdad",
}


@dataclass
class DesignChain:
    # Chain type: protein/dna/rna/ligand
    type: DesignChainType
    # Chain ID(s) for this chain in the PDB file
    # One chain or multiple chain IDs in case of multiple identical chains
    chain_ids: list[str]
    # Sequence (amino acid, DNA or RNA)
    sequence: str = None
    # SMILES string to define a ligand molecule
    smiles: str = None
    # CCD string to define a ligand molecule
    ccd: str = None
    # Contig string for this chain compatible with RFdiffusion motif scaffolding
    # Chains correspond to chains in input structure
    # May or may not contain the chain break suffix (/0)
    # Example: C123-456/10-10/C567-890/20-20/0
    contig: str = None
    # Macrocyclic chain
    cyclic: bool = False


@dataclass
class DesignSpec:
    chains: list[DesignChain]

    def get_chain(self, chain_id: str):
        for chain in self.chains:
            if chain_id in chain.chain_ids:
                return chain
        raise KeyError(f"Chain {chain_id} not found in design spec")

    @classmethod
    def from_pdb_str(cls, pdb_data: str, chains: list[str], cyclic=False):
        sequences = get_sequences_from_pdb_str(pdb_data, chains=chains)
        remarks = get_standardized_remarks_from_pdb_str(pdb_data)
        contig_by_chain = {}
        if remarks and remarks.get("Chains") and remarks.get("Standardized contig"):
            contig_chains = remarks["Standardized contig"].split()
            chains_list = remarks["Chains"].split()
            assert len(chains_list) == len(contig_chains)
            contig_by_chain = dict(zip(chains_list, contig_chains))
        return cls(
            chains=[
                DesignChain(
                    # TODO assuming all chains are protein, add support for ligands, dna, rna
                    type="protein",
                    # TODO enable detecting symmetric chains and using multiple chain ids here
                    chain_ids=[chain_id],
                    sequence=sequence,
                    contig=contig_by_chain.get(chain_id),
                    cyclic=cyclic,
                )
                for chain_id, sequence in sequences.items()
            ]
        )

    @classmethod
    def from_dict(cls, data):
        chains = [DesignChain(**chain_data) for chain_data in data.pop("chains")]
        return cls(chains=chains, **data)


class Design(Base):
    __tablename__ = "design"

    id: Mapped[str] = mapped_column(String, primary_key=True)
    pool_id: Mapped[str] = mapped_column(String, nullable=False, index=True)
    structure_path: Mapped[str] = mapped_column(String, default=None, nullable=True, info=dict(storage_path=True))
    structure_descriptor_key: Mapped[str] = mapped_column(String, default=None, nullable=True)
    accepted: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)
    spec: Mapped[DesignSpec] = mapped_column(DataclassType(DesignSpec), default=None, nullable=False)
    # This should be used when aligning design to its input - in case of multiple inputs
    contig_index: Mapped[int] = mapped_column(Integer, default=0, nullable=False)

    def __post_init__(self):
        assert self.pool_id, "pool_id must be set for Design"
        assert "_" not in self.pool_id, f"pool_id cannot contain underscores, got: {self.pool_id}"
        assert self.id, "id must be set for Design"
        if not self.id.startswith(f"ovo_{self.pool_id}_"):
            raise ValueError(f"Design ID '{self.id}' from pool '{self.pool_id}' must start with 'ovo_{self.pool_id}_'")

    @classmethod
    def design_id_to_pool_id(cls, design_id: str) -> str:
        """Get Pool ID from Design ID (ovo_xyz_1234 -> xyz)"""
        assert design_id.startswith("ovo_"), f"Invalid design ID '{design_id}'"
        assert design_id.count("_") >= 2, f"Invalid design ID '{design_id}'"
        return design_id.split("_")[1]

    @classmethod
    def from_pdb_file(
        cls,
        storage,
        filename: str,
        pdb_str: str,
        chains: list[str],
        project_id: str,
        pool_id: str,
        id: str = None,
        **kwargs,
    ):
        basename, file_extension = os.path.splitext(filename)
        if id is None:
            id = f"ovo_{pool_id}_{basename}"

        try:
            spec = DesignSpec.from_pdb_str(pdb_str, chains=chains)
        except ChainNotFoundError as e:
            raise ValueError(f"File '{filename}' does not contain requested chain '{e}'") from e

        return cls(
            id=id,
            pool_id=pool_id,
            spec=spec,
            structure_path=storage.store_file_str(
                pdb_str, os.path.join("project", project_id, "pools", pool_id, "designs", f"{id}.pdb")
            ),
            **kwargs,
        )


class DescriptorJob(Base, MetadataMixin, JobMixin):
    __tablename__ = "descriptor_job"

    id: Mapped[str] = mapped_column(String, primary_key=True, default=None)
    project_id: Mapped[str] = mapped_column(String, default=None, nullable=False, index=True)
    processed: Mapped[bool] = mapped_column(Boolean, default=False, nullable=False, index=True)
    workflow: Mapped[DescriptorWorkflow] = mapped_column(DataclassType(Workflow), default=None, nullable=False)

    @classmethod
    def generate_id(cls, tries=100):
        """Generate unique UID shortened for memory efficiency in descriptor_value table"""
        for length in range(6, 10):
            for i in range(tries):
                new_id = str(uuid.uuid4()).replace("-", "")[:length]
                try:
                    DescriptorJob.get(id=new_id)
                except NoResultFound:
                    return new_id
        raise ValueError(f"Failed to generate unique id of length {length} after {tries} tries")


class DescriptorValue(Base):
    __tablename__ = "descriptor_value"

    design_id: Mapped[str] = mapped_column(String, primary_key=True)
    descriptor_key: Mapped[str] = mapped_column(String, primary_key=True)
    descriptor_job_id: Mapped[str] = mapped_column(String, primary_key=True)
    # Comma-separated list of chain IDs for which this descriptor value applies
    # Can be same or a subset of descriptor_job.workflow.chains
    chains: Mapped[str] = mapped_column(String, primary_key=True)
    value: Mapped[str] = mapped_column(String, nullable=True)


class ArtifactTypes:
    # artifact type -> class
    _registry = {}

    @classmethod
    def register(cls):
        """Decorator to register an Artifact class"""

        def decorator(registered_class):
            artifact_type = f"{registered_class.__module__}.{registered_class.__qualname__}"
            cls._registry[artifact_type] = registered_class
            registered_class.artifact_type = artifact_type
            return registered_class

        return decorator

    @classmethod
    def exists(cls, artifact_type):
        """Check if an artifact of the given type exists"""
        return artifact_type in cls._registry

    @classmethod
    def get(cls, artifact_type):
        """Get artifact class by type string"""
        if artifact_type not in cls._registry:
            raise ValueError(
                f"Artifact type '{artifact_type}' is not registered, available types: {list(cls._registry.keys())}"
            )
        return cls._registry[artifact_type]


@dataclass
class Artifact(ABC):
    # artifact_type is filled in by @ArtifactTypes.register decorator
    artifact_type: str = field(init=False)

    @classmethod
    def from_dict(cls, data: dict) -> "Artifact":
        data_copy = copy(data)
        artifact_type: str = data_copy.pop("artifact_type")
        if not ArtifactTypes.exists(artifact_type):
            raise ValueError(f"Artifact type '{artifact_type}' is not registered")
        ArtifactSubclass = ArtifactTypes.get(artifact_type)
        return ArtifactSubclass(**data_copy)

    @abstractmethod
    def get_storage_paths(self) -> list[str]:
        raise NotImplementedError()


class ProjectArtifact(Base):
    __tablename__ = "project_artifact"
    id: Mapped[str] = mapped_column(String, primary_key=True, default_factory=lambda: str(uuid.uuid4()))
    project_id: Mapped[str] = mapped_column(String, nullable=False, default=None, index=True)
    artifact_type: Mapped[str] = mapped_column(String, nullable=False, default=None, index=True)
    descriptor_job_id: Mapped[str] = mapped_column(String, nullable=True, default=None, index=True)
    design_job_id: Mapped[str] = mapped_column(String, nullable=True, default=None, index=True)
    artifact: Mapped[Artifact] = mapped_column(DataclassType(Artifact), default=None, nullable=False)


@dataclass
class Descriptor(ABC):
    # Human-readable name of the descriptor
    name: str = None
    # Description of the descriptor
    description: str = None
    # Tool name (seq_composition, esm_1v,...)
    tool: str = None
    # Short machine-readable column name for the descriptor (sequence_composition|length, esm1v|native_seq_avg_softmax, ...)
    key: str = None

    def serialize(self, value):
        """Serialize a value into a number or string so that it can be stored in the DB"""
        return value

    def deserialize(self, value):
        """De-serialize a value stored in the DB into a meaningful object in python"""
        return value


@dataclass
class NumericDescriptor(Descriptor, ABC):
    # Comparison type (higher_is_better, lower_is_better, does_not_apply)
    comparison: str = None
    # Minimum value for the descriptor
    min_value: float = None
    # Maximum value for the descriptor
    max_value: float = None
    # Value above/below which the descriptor is considered a warning
    warning_value: float = None
    # Value above/below which the descriptor is considered an error
    error_value: float = None
    # Unit, e.g. "Å", "Å²", "kcal/mol"
    unit: str = None
    # Predefined color scale - see get_descriptor_cmap() function
    color_scale: str = None

    def get_plot_range(self, values: pd.Series, padding=0.05) -> tuple[float, float] | None:
        """Return range for the plot based on min_value, max_value with additional padding"""
        min_value = self.min_value if self.min_value is not None else values.min()
        if not pd.isna(values.min()) and values.min() < min_value:
            min_value = values.min()
        max_value = self.max_value if self.max_value is not None else values.max()
        if not pd.isna(values.max()) and values.max() > max_value:
            max_value = values.max()
        padding_value = (max_value - min_value) * padding
        return min_value - padding_value, max_value + padding_value

    def format(self, value):
        if pd.isna(value):
            return None
        unit_suffix = f" {self.unit}" if self.unit else ""
        try:
            if int(value) == value:
                return str(int(value)) + unit_suffix
        except:
            pass
        if isinstance(value, str):
            return value
        return f"{value:.2f}" + unit_suffix


@dataclass
class NumericGlobalDescriptor(NumericDescriptor):
    """Descriptor storing a number for the whole molecule"""

    pass


@dataclass
class NumericResidueDescriptor(NumericDescriptor):
    """Descriptor storing a number for each residue in a protein chain"""

    pass


@dataclass
class StringGlobalDescriptor(Descriptor):
    """Descriptor storing a string label for the whole molecule"""

    pass


@dataclass
class StringResidueDescriptor(Descriptor):
    """Descriptor storing a string for multiple residues, as a {resid -> string} dictionary,
    stored in the DB as a JSON-encoded string."""

    def serialize(self, value):
        """Serialize a value into a number or string so that it can be stored in the DB"""
        assert isinstance(value, dict), (
            f"Expected dict when serializing {self.name}, got value {value} ({type(value).__name__})"
        )
        return json.dumps(value)

    def deserialize(self, value):
        """De-serialize a value stored in the DB into a meaningful object in python"""
        return json.loads(value)


@dataclass
class CategoricalResidueDescriptor(Descriptor):
    """Descriptor storing a category label for each residue in a protein chain,
    each category represented by a letter.

    Annotation for the whole chain is saved as a single string,
    each category represented by a letter (key in category_labels),
    for example "EEEHHH---HHH" for DSSP secondary structure annotation.
    """

    # map from letter to category label (H -> Helix)
    category_labels: dict[str] = field(default_factory=dict)


@dataclass
class ResidueNumberDescriptor(Descriptor):
    """Descriptor storing list of residue numbers as a value, for example A123,A124,A202"""

    pass


@dataclass
class FileDescriptor(Descriptor):
    """Descriptor storing a Storage file path as a value"""

    pass


@dataclass
class StructureFileDescriptor(FileDescriptor):
    """Descriptor storing a structure file path as a value (PDB, mmCIF)"""

    structure_type: Literal["backbone_design", "sequence_design", "prediction", "experimentally_resolved"] = None
    b_factor_value: Literal["plddt", "fractional_plddt"] = None


# include all models
__all__ = [
    "Base",
    "JobMixin",
    "MetadataMixin",
    "UserSettings",
    "Project",
    "Round",
    "WorkflowParams",
    "Threshold",
    "WorkflowTypes",
    "Workflow",
    "DescriptorWorkflow",
    "DesignDescriptorWorkflow",
    "DesignWorkflow",
    "UnknownWorkflow",
    "DesignJob",
    "Pool",
    "DesignChainType",
    "DesignChain",
    "DesignSpec",
    "Design",
    "DescriptorJob",
    "DescriptorValue",
    "Descriptor",
    "NumericDescriptor",
    "NumericGlobalDescriptor",
    "NumericResidueDescriptor",
    "StringGlobalDescriptor",
    "StringResidueDescriptor",
    "CategoricalResidueDescriptor",
    "ResidueNumberDescriptor",
    "FileDescriptor",
    "StructureFileDescriptor",
    "DataclassType",
    "Artifact",
    "ArtifactTypes",
    "ProjectArtifact",
]



import traceback

from biopandas.pdb import PandasPdb
import pandas as pd
import numpy as np
import glob
from Bio.PDB import PDBParser, PDBIO
from Bio.PDB.SASA import ShrakeRupley
from scipy.spatial import distance
from biotite.structure import CellList
import argparse

import os
from io import StringIO
from utils import PDBSelector, get_resname_resnum_atomname_atomnum_str
import re
import pydssp_numpy

# Author: Christopher Warren, Marco Ancona, David Prihoda

pdb_parser: PDBParser = PDBParser(QUIET=True)

BACKBONE_CONTACT_DISTANCE = 10.0
SIDECHAIN_CONTACT_DISTANCE = 3.0


def get_all_metrics(
    pdbs: list | str, hotspots: list[int] | None = None, only_CA: bool = True, cyclic: bool = False
) -> dict[str, dict[str, float]]:
    """Calculate all metrics for a list of PDB files.

    :parameter:
        pdbs (list | str): List of PDB files or a single PDB file path.
        hotspots (list[int] | None): List of hotspot residues. Default: None.
        only_CA (bool): Use only CA atoms. Default: True.
    :returns:
        dict[str, dict[str, str]]: Dictionary with the calculated metrics.
    """

    if isinstance(pdbs, str):
        pdbs = [pdbs]

    scores = {}

    threshold_distance = BACKBONE_CONTACT_DISTANCE if only_CA else SIDECHAIN_CONTACT_DISTANCE

    for pdb in pdbs:
        glycosolated_target_residues = None
        glycosolated_binder_residues = None

        distances_interface_df = None
        average_interface_dist = None
        average_dist_interface_only_contacts = None
        N_contact_interface = 0

        distances_hotspots_df = None
        average_hotspots_dist = None
        average_dist_hotspots_only_contacts = None
        N_contact_hotspots = 0

        # Calculate all metrics
        pdb_name = os.path.basename(pdb).removesuffix(".pdb")

        # Read atom dataframe using BioPandas
        atom_ppdb = PandasPdb().read_pdb(pdb).df["ATOM"]

        # TODO we should parse this from the PDB standardized file header instead
        chain_ids = sorted(atom_ppdb.chain_id.unique())
        if chain_ids == ["A", "B"]:
            binder_chain, target_chain = chain_ids
        elif chain_ids == ["A"]:
            binder_chain, target_chain = chain_ids[0], None
        else:
            raise ValueError(f"Expected chains A and B, found {chain_ids} in {pdb}")

        Rg = radius_of_gyration(atom_ppdb, chain_id=binder_chain)

        glycosolated_binder_residues = glycosylation_sites(atom_ppdb, chain_id=binder_chain, cyclic=cyclic)

        scores[pdb_name] = {
            "radius_of_gyration": Rg,
            "glycosolated_binder_residues": glycosolated_binder_residues,
        }

        if target_chain:
            glycosolated_target_residues = glycosylation_sites(atom_ppdb, chain_id=target_chain)

            target_interfaces_str, binder_interfaces_str, binder_res_str = get_interface(
                pdb, threshold_distance, target_chain=target_chain, binder_chain=binder_chain, only_CA=only_CA
            )

            target_interface_residues = (
                [int(res_str.split("_")[1]) for res_str in target_interfaces_str.split(",")]
                if target_interfaces_str
                else []
            )
            # binder_interface_residues = [int(res_str.split('_')[1]) for res_str in binder_interfaces_str.split(',')] \
            #     if binder_interfaces_str else None
            binder_residues = (
                [int(res_str.split("_")[1]) for res_str in binder_res_str.split(",")] if binder_interfaces_str else []
            )

            if target_interface_residues and binder_residues:
                (
                    distances_interface_df,
                    average_interface_dist,
                    average_dist_interface_only_contacts,
                    N_contact_interface,
                ) = distance_matrix_binder_target_interface(
                    pdb,
                    target_chain=target_chain,
                    binder_chain=binder_chain,
                    target_interface_residues=target_interface_residues,
                    binder_interface_residues=binder_residues,
                    only_CA=only_CA,
                )

            scores[pdb_name].update(
                {
                    "glycosolated_target_residues": glycosolated_target_residues,
                    "interface_target_residues": ",".join([f"{target_chain}{r}" for r in target_interface_residues]),
                    "interface_binder_residues": ",".join([f"{binder_chain}{r}" for r in binder_residues]),
                    # "distances_binder_interface_against_target_interface": distances_interface_df,
                    # "average_dist_binder_to_target_interface": average_interface_dist,
                    # "average_dist_binder_to_target_interface_contacts": average_dist_interface_only_contacts,
                }
            )

            if hotspots:
                if binder_residues:
                    (
                        distances_hotspots_df,
                        average_hotspots_dist,
                        average_dist_hotspots_only_contacts,
                        N_contact_hotspots,
                    ) = distance_matrix_binder_target_interface(
                        pdb,
                        target_chain=target_chain,
                        binder_chain=binder_chain,
                        target_interface_residues=hotspots,
                        binder_interface_residues=binder_residues,
                        only_CA=only_CA,
                    )
                hotspots_on_interface = (
                    sorted(set(hotspots) & set(target_interface_residues))
                    if hotspots and target_interface_residues
                    else []
                )
                scores[pdb_name].update(
                    {
                        # "average_dist_binder_to_hotspots_contacts": average_dist_hotspots_only_contacts,
                        # "distances_binder_interface_against_hotspots": distances_hotspots_df,
                        "interface_target_hotspot_residues": ",".join(
                            [f"{target_chain}{r}" for r in hotspots_on_interface]
                        ),
                        "N_contact_interface": N_contact_interface,
                        "average_dist_binder_to_hotspots": average_hotspots_dist,
                        "N_hotspots_on_interface": len(hotspots_on_interface),
                        "N_contact_hotspots": N_contact_hotspots,
                    }
                )

        try:
            pydssp_str = pydssp_assign(pdb, chain_id=binder_chain, cyclic=cyclic)
            scores[pdb_name].update(
                {
                    "pydssp_str": pydssp_str,  # DSSP line for the PDB file (H = alpha-helix, E = beta-strand, - = loop)
                    "pydssp_loop_percent": pydssp_str.count("-") / len(pydssp_str) * 100,
                    "pydssp_helix_percent": pydssp_str.count("H") / len(pydssp_str) * 100,
                    "pydssp_strand_percent": pydssp_str.count("E") / len(pydssp_str) * 100,
                }
            )
        except Exception as e:
            traceback.print_exc()
            scores[pdb_name]["pydssp_error"] = f"Unexpected pydssp error: {e}"

    return scores


# TODO: add this in the utils.py file
def get_coords_from_ppdb(atom_ppdb: pd.DataFrame) -> np.ndarray | None:
    """Get coordinates from a PandasPdb DataFrame.

    :parameter:
        atom_ppdb (pd.DataFrame): Pandas DataFrame with ATOM records.
    :returns:
        np.ndarray: NumPy array with coordinates.
    """

    if atom_ppdb.empty:
        return None

    assert atom_ppdb.record_name.unique() == ["ATOM"], "Only ATOM records are allowed when getting coordinates."

    return atom_ppdb[["x_coord", "y_coord", "z_coord"]].values


def radius_of_gyration(ppdb_atoms, chain_id: str | None, only_CA: bool = True) -> float | None:
    """Calculate radius of gyration of polymer, i.e. compactness score of a protein structure.

    :parameter:
        pdb_path: PDB file path.
        chain_id (str | None): Chain ID to use, or None to use all chains
    :returns:
        float: Compactness score.
    """

    if only_CA:
        ppdb_atoms = ppdb_atoms.query("atom_name == 'CA'")

    if chain_id is not None:
        ppdb_atoms = ppdb_atoms.query("chain_id==@chain_id")

    coords = get_coords_from_ppdb(ppdb_atoms)

    if coords is None:
        return None

    center_of_mass = coords.mean(axis=0)
    squared_distances = np.sum((coords - center_of_mass) ** 2, axis=1)
    Rg = np.sqrt(np.mean(squared_distances))

    return float(Rg)


def glycosylation_sites(atom_ppdb: pd.DataFrame, chain_id: str, cyclic: bool = False) -> str | None:
    """Find glycosylation sites in a PDB file. Calls detect_glycosylation_sites

    :returns:
        str: Comma-separated list of glycosylated residue numbers. None if no glycosylation sites are found.
    """

    glycosylation_dict = detect_glycosylation_sites(atom_ppdb, chains=chain_id, cyclic=cyclic)
    if not glycosylation_dict:
        return None

    return ",".join([str(resnum) for (chain, resnum) in glycosylation_dict])


def detect_glycosylation_sites(
    atom_ppdb: pd.DataFrame,
    chains: list[str] | str | None = None,
    query_atoms: list[str] | None = None,
    cyclic: bool = False,
) -> dict | None:
    """Find glycosylation sites in the PDB file
    :param:
        atom_ppdb: Pandas DataFrame with the ATOM records of the PDB file
        chains: str or list of str with the chain IDs to search for glycosylation sites
        query_atoms : list of str with the atom names to search for glycosylation sites

    :return:
        glycosylation_dict: dictionary with the coordinates of the query glycosylated atoms
    """
    glycosylation_pattern = ["ASN", "* except PRO", ["THR", "SER"]]  # Define the pattern N*S N*T

    if query_atoms is None:
        query_atoms = ["CA", "C", "CB", "ND2"]

    if isinstance(chains, str):
        chains = [chains]
    elif chains is None:
        chains = atom_ppdb["chain_id"].unique()

    atom_ppdb = atom_ppdb.query("chain_id in @chains")

    residues = atom_ppdb[["chain_id", "residue_number", "residue_name"]].drop_duplicates().set_index("residue_number")

    # Create a DataFrame of overlapping residue triplets (e.g. resnum
    # Before dropna three_mers looks like this:
    # num chain res[num] res[num+1] res[num+2]
    # 129   A     ASN       THR        THR
    # 130   A     THR       THR        VAL
    # 131   A     THR       VAL        PHE
    # 132   A     VAL       PHE        None
    # 132   A     PHE       None       None

    three_mers = pd.concat(
        [residues.chain_id] + [residues.residue_name.shift(-i) for i in range(len(glycosylation_pattern))], axis=1
    )

    three_mers.columns = ["chain_id", "resname1", "resname2", "resname3"]

    if cyclic:
        three_mers.loc[three_mers.index[-1], "resname2"] = three_mers.iloc[0].loc["resname1"]
        three_mers.loc[three_mers.index[-1], "resname3"] = three_mers.iloc[0].loc["resname2"]
        three_mers.loc[three_mers.index[-2], "resname3"] = three_mers.iloc[0].loc["resname1"]

    three_mers = three_mers.dropna()

    # Check for the glycosylation pattern
    is_glycosylated = three_mers.apply(
        lambda row: (row.iloc[1] == glycosylation_pattern[0])
        and (row.iloc[2] != "PRO")
        and (row.iloc[3] in glycosylation_pattern[2]),
        axis=1,
    )

    glycosylated_residues = three_mers[is_glycosylated]

    if glycosylated_residues.empty:
        return None

    glycosylated_chains = glycosylated_residues.chain_id.unique()
    glycosylation_dict = {}
    for chain in glycosylated_chains:
        chain_atom_ppdb = atom_ppdb.query("chain_id==@chain")
        for resnum in glycosylated_residues[glycosylated_residues.chain_id == chain].index:
            glycosylation_dict[(chain, resnum)] = {}
            for atom in query_atoms:
                res_coords_df = chain_atom_ppdb.query("residue_number==@resnum")
                glycosylation_dict[(chain, resnum)][atom] = res_coords_df.query("atom_name==@atom")[
                    ["x_coord", "y_coord", "z_coord"]
                ].values.flatten()

    return glycosylation_dict


def calculate_sasa_per_chain(struc, chain, **args) -> pd.Series:
    """Calculate the solvent-accessible surface area (SASA) for a chain in a PDB structure."""

    # This is the most time-consuming calculation
    sr = ShrakeRupley(**args)

    sr.compute(struc, level="R")

    sasa = {}
    for residue in struc[0][chain]:
        sasa[residue.id] = residue.sasa

    return pd.Series(sasa)


def DeltaSASA(complex_pdb_file: str, chain: str, **args) -> pd.DataFrame:
    """Calculate the DeltaSASA for a chain in a complex PDB file.
    :parameter:
        complex_pdb_file (str): Path to the PDB file.
        chain (str): Chain ID.
    :returns:
        pd.DataFrame: DataFrame with the DeltaSASA values.
    """

    parser = PDBParser(QUIET=True)
    complex_structure = parser.get_structure("complex", complex_pdb_file)

    io = PDBIO()
    io.set_structure(complex_structure)

    target_pdb_file = StringIO()
    io.save(target_pdb_file, select=PDBSelector(chain))

    target_structure = parser.get_structure("target", StringIO(target_pdb_file.getvalue()))

    # Calculate SASA for the entire complex
    complex_sasa = calculate_sasa_per_chain(complex_structure, chain, **args)
    target_sasa = calculate_sasa_per_chain(target_structure, chain, **args)

    interface_deltaSASA = pd.concat(
        [
            pd.Series([res.resname for res in target_structure[0][chain]], index=complex_sasa.index, name="residue"),
            complex_sasa.rename("complex_SASA"),
            target_sasa.rename("target_SASA"),
            (complex_sasa - target_sasa).rename("DeltaSASA"),
        ],
        axis=1,
    )

    return interface_deltaSASA


def get_interface(
    complex_pdb_file, threshold_distance, target_chain: str, binder_chain: str, only_CA=True
) -> tuple[str | None, str | None, str | None]:
    """Get the interface residues from a PDB file.

    :parameter:
        complex_pdb_file (str): Path to the PDB file.
        threshold_distance (float): Threshold distance to consider a contact.
        target_chain: ID of the target chain (usually B).
        binder_chain: ID of the binder chain (usually A).
        only_CA (bool): Use only CA atoms. Default: True.
    :returns:
        Tuple with the target (comma-separated list) residues and binder interfaces (comma-separated list) residues.
        In the format: RESNAME_RESNUM_ATOMNAME_ATOMNUM
    """
    ppdb = PandasPdb().read_pdb(complex_pdb_file)
    atom_ppdb = ppdb.df["ATOM"]
    unique_chains = set(atom_ppdb.chain_id.unique())
    assert unique_chains == {target_chain, binder_chain}, (
        f"Only target ({target_chain}) and binder ({binder_chain}) chains should be found in PDB, "
        f"found: {unique_chains}"
    )

    if only_CA:
        atom_ppdb = atom_ppdb[atom_ppdb.atom_name == "CA"]

    indices_binder = atom_ppdb.query("chain_id == @binder_chain").index
    indices_target = atom_ppdb.query("chain_id == @target_chain").index

    matrix = CellList(get_coords_from_ppdb(atom_ppdb), threshold_distance).create_adjacency_matrix(threshold_distance)
    matrix = pd.DataFrame(matrix, index=atom_ppdb.index, columns=atom_ppdb.index).loc[indices_target, indices_binder]

    # Select interacting matrix
    interacting_matrix = matrix.loc[matrix.any(axis=1), matrix.any(axis=0)]

    if interacting_matrix.empty:
        return None, None, None

    # Select target and binder interfaces residues (RESNAME_RESNUM_ATOMNAME_ATOMNUM)
    target_interfaces_str = get_resname_resnum_atomname_atomnum_str(atom_ppdb, list(interacting_matrix.index))
    binder_interfaces_str = get_resname_resnum_atomname_atomnum_str(atom_ppdb, list(interacting_matrix.columns))
    binder_str = get_resname_resnum_atomname_atomnum_str(atom_ppdb, list(indices_binder))

    return target_interfaces_str, binder_interfaces_str, binder_str


def distance_matrix_binder_target_interface(
    complex_pdb_file,
    target_chain: str,  # used in pandas query
    binder_chain: str,  # used in pandas query
    target_interface_residues: list[int],
    binder_interface_residues: list[int],
    only_CA: bool = True,
) -> tuple[pd.DataFrame | None, float | None, float | None, int | None]:
    """Calculate the distance matrix between the target and binder interfaces in a PDB file (complex).

    :parameter:
        complex_pdb_file (str): Path to the PDB file.
        target_interface_residues (list[int]): List of target interface residues.
        binder_interface_residues (list[int]): List of binder interface residues.
        target_chain: ID of the target chain (usually B).
        binder_chain: ID of the binder chain (usually A).
        only_CA (bool): Use only CA atoms. Default: True.
    :returns:
        Tuple with the distance matrix, average distance, average distance and number of contacts.
    """

    ppdb = PandasPdb().read_pdb(complex_pdb_file)

    assert target_interface_residues is not None, (
        f"Target interface residues must be provided, found: {target_interface_residues}"
    )
    assert binder_interface_residues is not None, (
        f"Target interface residues must be provided, found: {binder_interface_residues}"
    )

    target_interface = ppdb.df["ATOM"].query(
        "(residue_number in @target_interface_residues) & (chain_id == @target_chain)"
    )
    binder_interface = ppdb.df["ATOM"].query(
        "(residue_number in @binder_interface_residues) & (chain_id == @binder_chain)"
    )

    if only_CA:
        target_interface = target_interface.query("atom_name=='CA'")
        binder_interface = binder_interface.query("atom_name=='CA'")

    coords_target_interface = get_coords_from_ppdb(target_interface)
    coords_binder_interface = get_coords_from_ppdb(binder_interface)

    if coords_target_interface is None or coords_binder_interface is None:
        return None, None, None, 0

    # Calculate pairwise distances
    distances = distance.cdist(coords_target_interface, coords_binder_interface, metric="euclidean")

    index_target = target_interface.apply(
        lambda row: "_".join([row.residue_name, str(row.residue_number), row.atom_name, str(row.atom_number)]), axis=1
    )
    index_binder = binder_interface.apply(
        lambda row: "_".join([row.residue_name, str(row.residue_number), row.atom_name, str(row.atom_number)]), axis=1
    )

    distances_df = pd.DataFrame(distances, index=index_target, columns=index_binder)

    average_dist = distances_df.mean().mean()
    contact_distance = BACKBONE_CONTACT_DISTANCE if only_CA else SIDECHAIN_CONTACT_DISTANCE
    N_contacts = (distances < contact_distance).sum() if len(distances) else 0

    average_dist_only_contacts = distances[distances < contact_distance].mean() if N_contacts else None

    return distances_df, average_dist, average_dist_only_contacts, N_contacts


def pydssp_assign(pdb_path, chain_id: str | None, cyclic=False, tail_residues=5) -> str:
    """Return DSSP line for a PDB file using pydssp_numpy (- = loop, H = alpha-helix, E = beta-strand)"""
    C3_ALPHABET = np.array(["-", "H", "E"])
    coord = pydssp_numpy.read_pdbtext_with_checking(open(pdb_path, "r").read(), chain_id=chain_id)
    tail_residues = min(tail_residues, len(coord))
    if cyclic:
        coord = np.concatenate([coord[-tail_residues:], coord, coord[:tail_residues]], axis=0)
    # main calculation
    onehot = pydssp_numpy.assign(coord)
    if cyclic:
        onehot = onehot[tail_residues:-tail_residues]
    index = np.argmax(onehot, axis=-1)
    return "".join(list(C3_ALPHABET[index]))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_path", type=str)
    parser.add_argument("output_csv", type=str)
    parser.add_argument("--hotspot", type=str, default=None, help="Hotspot chain-resnum(s), comma-separated")
    parser.add_argument(
        "--cyclic", action="store_true", default=False, help="Connect C and N term to form a macrocyclic peptide"
    )
    parser.add_argument(
        "--filters", type=str, default="none", help="Filters, e.g. 'N_contact_hotspots>5,pydssp_helix_percent<50'"
    )
    parser.add_argument("--filtered-output", type=str, default=None, help="Directory path to save filtered PDBs")

    options = parser.parse_args()

    if options.hotspot:
        chains, hotspots = zip(*[(chain_resnum[0], chain_resnum[1:]) for chain_resnum in options.hotspot.split(",")])
        assert set(chains) == {"B"}, (
            f"Only target chain 'B' hotspots allowed, found prefix: {set(chains)} in hotspots {options.hotspot}"
        )
        hotspots = [int(hotspot) for hotspot in hotspots]
    else:
        hotspots = None

    if os.path.isdir(options.input_path):
        paths = sorted(glob.glob(os.path.join(options.input_path, "*.pdb")))
        print(f"Reading sequences from {len(paths):,} PDBs")
    elif options.input_path.endswith((".pdb",)):
        paths = options.input_path
    else:
        raise ValueError("Input must be a directory with PDB files, or a PDB file.")

    parsed_filters = []
    for filter_str in options.filters.split(","):
        if filter_str == "none":
            continue
        comp = None
        for c in [">=", "<=", ">", "<", "="]:
            if c in filter_str:
                comp = c
                break
        if not comp:
            raise ValueError(f"Invalid filter: {filter_str}")
        field, value = filter_str.split(comp)
        field = field.strip()
        value = value.strip()
        try:
            value = float(value)
        except ValueError:
            pass
        parsed_filters.append((field, comp, value))

    if parsed_filters:
        assert options.filtered_output, "--filtered-output directory must be provided when using filters"

    scores = get_all_metrics(paths, hotspots=hotspots, cyclic=options.cyclic)

    df = pd.DataFrame(scores).T.rename_axis("id")

    if options.filtered_output:
        if not os.path.exists(options.filtered_output):
            os.makedirs(options.filtered_output)
        passed_per_design = {}
        passed_per_filter_per_design = {}
        for path in paths:
            pdb_name = os.path.basename(path).removesuffix(".pdb")
            passed = True
            if parsed_filters:
                row = df.loc[pdb_name]
                for field, comp, value in parsed_filters:
                    passed_field = True
                    if field not in row.index:
                        raise ValueError(f"Filter field {field} not found in metrics: {', '.join(row.index.tolist())}")
                    if comp == ">":
                        if pd.isna(row[field]) or not row[field] > value:
                            passed_field = False
                    elif comp == "<":
                        if pd.isna(row[field]) or not row[field] < value:
                            passed_field = False
                    elif comp == ">=":
                        if pd.isna(row[field]) or not row[field] >= value:
                            passed_field = False
                    elif comp == "<=":
                        if pd.isna(row[field]) or not row[field] <= value:
                            passed_field = False
                    elif comp == "=":
                        if not row[field] == value:
                            passed_field = False
                    else:
                        raise ValueError(f"Invalid comparison operator: {comp}")

                    if not passed_field:
                        passed = False

                    key = f"{field}{comp}{value}"
                    if key not in passed_per_filter_per_design:
                        passed_per_filter_per_design[key] = {}
                    passed_per_filter_per_design[key][pdb_name] = passed_field

            passed_per_design[pdb_name] = passed
            if passed:
                # link path
                os.symlink(os.path.abspath(path), os.path.join(options.filtered_output, pdb_name + ".pdb"))

        if parsed_filters:
            df["passed_filters"] = pd.Series(passed_per_design)
            print(f"all filters passed by {df.passed_filters.mean():.2%} designs")
            for key, vals in passed_per_filter_per_design.items():
                df[key] = pd.Series(vals)
                print(f"{key} passed by {df[key].mean():.2%} designs")
        df.to_csv(options.output_csv)
        print("Saved metrics to:", options.output_csv)
        print("Saved filtered PDBs to:", options.output_csv)



import streamlit as st
import os
from contigs_organizer import contigs_organizer

# `$ python -m streamlit run molstar_custom_component/main.py`
# do not forget to have venv activated

st.subheader("Component")

if os.path.exists("./test_files/multichain_more/result_0.pdb"):
    pdb = open("./test_files/multichain_more/result_0.pdb").read()
else:
    pdb = ""

co = contigs_organizer(
    contigs="22/A38-57/10/A76-94/10-20/A121-125", pdb=pdb, colors={}, key="test"
)  # parsing the full contig
# co = contigs_organizer(contigs="A38-57, A76-94, A115-120", liveInput=True, pdb=pdb) # parsing the fixed regions

st.write(co)

# contigs_organizer(contigs="A35-48/10/A107-123/0 B123-233")



import streamlit as st

from ovo import db
from ovo.app.components.descriptor_job_components import refresh_descriptors
from ovo.app.components.download_component import download_job_designs_component
from ovo.app.components.descriptor_scatterplot import (
    descriptor_scatterplot_design_explorer_component,
    descriptor_scatterplot_input_component,
)
from ovo.app.components.navigation import design_navigation_selector
from ovo.app.utils.cached_db import get_cached_design, get_cached_pool, get_cached_design_job
from ovo.core.database.models import Design, Pool, DesignWorkflow, WorkflowTypes, UnknownWorkflow
from ovo.app.utils.cached_db import get_cached_pools, get_cached_design_jobs


def explorer_fragment(pool_ids: list[str], design_ids: list[str] | None = None):
    pools = get_cached_pools(pool_ids)

    if design_ids is None:
        # design_ids not explicitly passed, use all accepted designs in the selected pools
        design_ids = sorted(db.select_values(Design, "id", pool_id__in=pool_ids, accepted=True))

        if not design_ids:
            st.write(
                "No accepted designs in the selected "
                + ("pools" if len(pool_ids) > 1 else "pool")
                + ". Please mark some designs as accepted in the **Jobs** page."
            )
            return

    elif not design_ids:
        # Empty list explicitly passed to design_ids
        st.write("No designs selected")
        return

    st.header(f"Explorer | {len(design_ids):,} {'design' if len(design_ids) == 1 else 'designs'}")

    badges = [f":grey-badge[{pool.name}]" for pool in pools]
    st.write("Selected pools: " + " ".join(badges))

    refresh_descriptors(
        design_ids=design_ids,
    )

    scatterplot_settings = descriptor_scatterplot_input_component(design_ids)
    if not scatterplot_settings:
        st.write("No descriptors available for the selected designs.")
        st.stop()

    selected_design_ids = descriptor_scatterplot_design_explorer_component(
        settings=scatterplot_settings, design_ids=design_ids
    )

    st.subheader("Download")
    download_job_designs_component(selected_design_ids, pools)

    st.subheader("Designs")
    design_visualization_fragment(selected_design_ids)


@st.fragment
def design_visualization_fragment(selected_design_ids: list[str]):
    # Use the navigation selector component
    design_id = design_navigation_selector(selected_design_ids, key="selected_design")

    st.subheader(design_id)

    design = get_cached_design(design_id)
    pool = get_cached_pool(design.pool_id)
    design_job = get_cached_design_job(pool.design_job_id) if pool.design_job_id else None
    WorkflowType = type(design_job.workflow) if design_job and design_job.workflow else DesignWorkflow

    if design_job and design_job.workflow.is_instance(UnknownWorkflow):
        st.warning(
            f"Falling back to basic structure and sequence visualization, "
            f"failed to load workflow information: {design_job.workflow.error}"
        )

    WorkflowType.visualize_single_design_structures(design_id)

    st.markdown("### Sequence")

    WorkflowType.visualize_single_design_sequences(design_id)

    st.markdown(f"### Download {design_id}")

    download_job_designs_component(design_ids=[design_id], pools=[pool], key="single")



from ovo.core.database.models import (
    NumericGlobalDescriptor,
    ResidueNumberDescriptor,
    StructureFileDescriptor,
    FileDescriptor,
    CategoricalResidueDescriptor,
    Descriptor,
)

# HOTSPOT INTERFACE METRICS
RADIUS_OF_GYRATION = NumericGlobalDescriptor(
    name="Radius of gyration (backbone)",
    description="Radius of gyration of the CA atoms of the protein",
    tool="Backbone metrics",
    key="rfd_ee|backbone_metrics|radius_of_gyration",
    comparison="lower_is_better",
)

N_CONTACTS_TO_INTERFACE = NumericGlobalDescriptor(
    name="N binder-target backbone contacts",
    description="Number of contacts between binder CA atoms and target CA atoms. "
    "Two CA atoms are in contact if they are within 8A from each other",
    tool="Backbone metrics",
    key="rfd_ee|backbone_metrics|N_contact_interface",
    min_value=0,
    comparison="higher_is_better",
)
INTERFACE_TARGET_RESIDUES = ResidueNumberDescriptor(
    name="Interface target residues",
    description="List of residue numbers of the target in contact with the binder backbone (CA within 8A)",
    tool="Backbone metrics",
    key="rfd_ee|backbone_metrics|interface_target_residues",
)

N_CONTACTS_TO_HOTSPOTS = NumericGlobalDescriptor(
    name="N binder-hotspots backbone contacts",
    description="Number of contacts between binder CA atoms and user-defined hotspots CA atoms. "
    "Two CA atoms are in contact if they are within 8A from each other",
    tool="Backbone metrics",
    key="rfd_ee|backbone_metrics|N_contact_hotspots",
    min_value=0,
    comparison="higher_is_better",
)

N_HOTSPOTS_ON_INTERFACE = NumericGlobalDescriptor(
    name="N hotspots on interface",
    description="Number of user-defined hotspot residues found on the binder-target backbone interface (CA within 8A)",
    tool="Backbone metrics",
    key="rfd_ee|backbone_metrics|N_hotspots_on_interface",
    min_value=0,
    comparison="does_not_apply",
)

PYDSSP_STRING = CategoricalResidueDescriptor(
    name="PyDSSP String",
    description="Secondary structure representation of the protein according to PyDSSP (H = helix, E = strand, - = loop)",
    tool="PyDSSP",
    key="rfd_ee|backbone_metrics|pydssp_str",
    category_labels={"H": "Helix", "E": "Strand", "-": "Loop"},
)

PYDSSP_LOOP_PERCENT = NumericGlobalDescriptor(
    name="PyDSSP Loop %",
    description="Percentage of residues with loop secondary structure according to PyDSSP",
    tool="PyDSSP",
    key="rfd_ee|backbone_metrics|pydssp_loop_percent",
    comparison="does_not_apply",
    min_value=0,
    max_value=100,
)

PYDSSP_HELIX_PERCENT = NumericGlobalDescriptor(
    name="PyDSSP Helix %",
    description="Percentage of residues with alpha-helix secondary structure according to PyDSSP",
    tool="PyDSSP",
    key="rfd_ee|backbone_metrics|pydssp_helix_percent",
    comparison="does_not_apply",
    min_value=0,
    max_value=100,
)

PYDSSP_SHEET_PERCENT = NumericGlobalDescriptor(
    name="PyDSSP Beta Sheet %",
    description="Percentage of residues with beta sheet secondary structure according to PyDSSP",
    tool="PyDSSP",
    key="rfd_ee|backbone_metrics|pydssp_strand_percent",
    comparison="does_not_apply",
    min_value=0,
    max_value=100,
)

BACKBONE_METRICS = [
    RADIUS_OF_GYRATION,
    N_CONTACTS_TO_INTERFACE,
    N_CONTACTS_TO_HOTSPOTS,
    N_HOTSPOTS_ON_INTERFACE,
    INTERFACE_TARGET_RESIDUES,
    PYDSSP_STRING,
    PYDSSP_LOOP_PERCENT,
    PYDSSP_HELIX_PERCENT,
    PYDSSP_SHEET_PERCENT,
]
BACKBONE_METRIC_FIELD_NAMES = [d.key.removeprefix("rfd_ee|backbone_metrics|") for d in BACKBONE_METRICS]


# AVERAGE_DISTANCE_BINDER_TO_TARGET_INTERFACE = NumericGlobalDescriptor(
#     name='Binder-interface average distance',
#     description='Mean distance of binder CA-atoms to the target interface CA atoms',
#     tool='Backbone metrics',
#     key='rfd_ee|backbone_metrics|average_dist_binder_to_target_interface',
#     min_value=0,
#     comparison='does_not_apply'
# )
#
# AVERAGE_DISTANCE_BINDER_TO_TARGET_INTERFACE_CONTACTS = NumericGlobalDescriptor(
#     name='Binder-interface contacts average distance',
#     description='Mean distance of binder CA-atoms to the target interface CA atoms which contact binder CA atoms (<8A)',
#     tool='Backbone metrics',
#     key='rfd_ee|backbone_metrics|average_dist_binder_to_target_interface_contacts',
#     min_value=0,
#     comparison='does_not_apply'
# )
#
# AVERAGE_DISTANCE_BINDER_TO_HOTSPOTS = NumericGlobalDescriptor(
#     name='Binder-hotspots average distance',
#     description='Mean distance of binder CA-atoms to the target hotspot CA atoms',
#     tool='Backbone metrics',
#     key='rfd_ee|backbone_metrics|average_dist_binder_to_hotspots',
#     min_value=0,
#     comparison='does_not_apply'
# )
#
# AVERAGE_DISTANCE_BINDER_TO_HOTSPOT_CONTACTS = NumericGlobalDescriptor(
#     name='Binder-hotspots contacts average distance',
#     description='Mean distance of binder CA-atoms to the target hotspot CA atoms which contact binder CA atoms (<8A)',
#     tool='Backbone metrics',
#     key='rfd_ee|backbone_metrics|average_dist_binder_to_hotspots_contacts',
#     min_value=0,
#     comparison='does_not_apply'
# )


RFDIFFUSION_STRUCTURE_PATH = StructureFileDescriptor(
    name="RFdiffusion backbone design",
    description="RFdiffusion-generated backbone structure, without any side-chains, with Glycine residues at designed positions",
    tool="RFdiffusion",
    key="rfd_ee|rfdiffusion|backbone_structure_path",
    structure_type="backbone_design",
)


RFDIFFUSION_TRB_PATH = FileDescriptor(
    name="RFdiffusion .trb file",
    description="RFdiffusion pickle file containing metadata about the designed backbone structure",
    tool="RFdiffusion",
    key="rfd_ee|rfdiffusion|backbone_trb_path",
)


LIGANDMPNN_STRUCTURE_PATH = StructureFileDescriptor(
    name="LigandMPNN sequence design",
    description="Structure for LigandMPNN-generated sequence with packed side-chains",
    tool="LigandMPNN",
    key="rfd_ee|ligandmpnn|sequence_structure_path",
    structure_type="sequence_design",
)


FASTRELAX_STRUCTURE_PATH = StructureFileDescriptor(
    name="ProteinMPNN FastRelax sequence design",
    description="Structure for sequence generated with ProteinMPNN and PyRosetta FastRelax protocol. Side-chains may or may not be relaxed depending on the cycle (last cycle is not relaxed in some implementations)",
    tool="ProteinMPNN+PyRosetta",
    key="rfd_ee|proteinmpnn_fastrelax|sequence_structure_path",
    structure_type="sequence_design",
)


# ROSETTA INTERFACE METRICS

# "contact_molecular_surface": 484.94976806640625,
# "ddg": -12.693643569946289,
# sap_score: 0.0,

PYROSETTA_DDG = NumericGlobalDescriptor(
    name="Rosetta ddG",
    description="Rosetta binding energy change (in Rosetta energy units, beta_nov16 scoring function). Lower (negative) values indicate better binding affinity.",
    tool="PyRosetta",
    key="pyrosetta_interface_metrics|pyrosetta|ddg",
    comparison="lower_is_better",
)

PYROSETTA_CMS = NumericGlobalDescriptor(
    name="Contact molecular surface",
    description="Contact molecular surface area (in Angstrom squared) calculated using Rosetta ContactMolecularSurface",
    unit="Å²",
    tool="PyRosetta",
    key="pyrosetta_interface_metrics|pyrosetta|contact_molecular_surface",
    comparison="higher_is_better",
)

PYROSETTA_SAP_SCORE = NumericGlobalDescriptor(
    name="Spatial aggregation propensity (SAP)",
    description="Hydrophobicity of surface exposed regions calculated using Rosetta SapScore. Positive values indicate hydrophobic molecules.",
    tool="PyRosetta",
    key="pyrosetta_interface_metrics|pyrosetta|sap_score",
    comparison="does_not_apply",
)

PYROSETTA_INTERFACE_DESCRIPTORS = [PYROSETTA_DDG, PYROSETTA_CMS, PYROSETTA_SAP_SCORE]

DESCRIPTORS = [v for v in globals().values() if isinstance(v, Descriptor)]

PRESETS = [
    {
        "label": "Rosetta ddG & CMS",
        "x": PYROSETTA_DDG,
        "y": PYROSETTA_CMS,
    },
    {
        "label": "PyDSSP Helix vs Sheet",
        "x": PYDSSP_HELIX_PERCENT,
        "y": PYDSSP_SHEET_PERCENT,
    },
]



import streamlit as st
import pandas as pd
from typing import List, Collection
from cmap import Colormap

from ovo.core.database.models import Descriptor, NumericGlobalDescriptor, ResidueNumberDescriptor
from ovo.core.logic.proteinqc_logic import get_descriptor_cmap


@st.cache_data(hash_funcs={Descriptor: lambda obj: (obj.key)})
def get_descriptor_column_config(descriptor: Descriptor):
    """
    Create a streamlit column_config for a more readable dataframe (labels, help, formatting,...)
    """
    if not isinstance(descriptor, NumericGlobalDescriptor):
        return st.column_config.TextColumn(
            label=descriptor.name, help=f"**{descriptor.name}**: {descriptor.description}", width="small"
        )
    if descriptor.name == "Sequence length":
        format = "%.0f"
    elif "%" in descriptor.name:
        format = "%.2f %%"
    else:
        format = "%.2f"
    return st.column_config.NumberColumn(
        label=descriptor.name, help=f"**{descriptor.name}**: {descriptor.description}", format=format, width="small"
    )


def get_residue_number_column_config(
    descriptor: Descriptor, unique_values: Collection, cmap_name: str = "seaborn:tab10_light", n_colors: int = 10
):
    if isinstance(descriptor, ResidueNumberDescriptor):
        cmap = Colormap(cmap_name)
        colors = list(color.hex for color in cmap.iter_colors(n_colors))
        return st.column_config.MultiselectColumn(
            label=descriptor.name,
            help=f"**{descriptor.name}**: {descriptor.description}",
            width="small",
            options=unique_values,
            color=colors,
        )
    return None


def make_bg_color_func(descriptor, min_val, max_val):
    cmap = get_descriptor_cmap(descriptor, min_val, max_val)

    def color_func(val):
        if pd.isna(val):
            return ""
        return f"background-color: {cmap(val)}"

    return color_func


@st.fragment()
def descriptor_table(design_ids: List[str], descriptors_df: pd.DataFrame, descriptors: List[Descriptor]):
    if not design_ids or descriptors_df.empty:
        st.warning("No results yet")
        return
    # Select only columns for the selected descriptors
    assert descriptors_df.columns.nlevels == 2, "Expected nested=True table with 2 levels (tool name, descriptor name)"
    assert len(set(d.key for d in descriptors)) == len(descriptors), (
        f"Duplicate descriptors found in list: {[d.key for d in descriptors]}"
    )
    sequence_cols = [col for col in descriptors_df.columns if col[0] == "Sequence"]
    selected_df = descriptors_df[sequence_cols + [(descriptor.tool, descriptor.name) for descriptor in descriptors]]
    styles = {}
    column_config = {}

    # When using positional column indices for column config, 0 refers to the first index column
    col_offset = selected_df.index.nlevels

    for descriptor in descriptors:
        col = (descriptor.tool, descriptor.name)
        if col not in selected_df.columns:
            continue
        col_idx = selected_df.columns.get_loc(col) + col_offset
        if isinstance(descriptor, NumericGlobalDescriptor):
            min_val, max_val = selected_df[col].agg(["min", "max"])
        else:
            min_val, max_val = None, None
        styles[col] = selected_df[col].map(make_bg_color_func(descriptor, min_val, max_val))
        if isinstance(descriptor, ResidueNumberDescriptor):
            unique_values = set(x.strip() for row in descriptors_df[col].dropna() for x in row.split(","))
            column_config[col_idx] = get_residue_number_column_config(descriptor, unique_values)
        else:
            column_config[col_idx] = get_descriptor_column_config(descriptor)

    # Combine styles into a DataFrame
    style_df = pd.DataFrame(styles)
    styled_df = selected_df.style.apply(lambda _: style_df, axis=None)

    st.dataframe(styled_df, column_config=column_config, key="descriptor_table")


def residue_number_descriptor_detail_table(descriptor: Descriptor, descriptor_values: pd.Series):
    """
    Returns (table_data, column_config, caption, format_func) for ResidueNumberDescriptor detail view.
    table_data: DataFrame with designs as rows, unique residues as columns, checkmark if present.
    column_config: dict for st.dataframe
    caption: str
    format_func: function for design_id formatting
    """
    unique_residues = set()
    design_to_residues = {}

    # Build mapping from design to set of residues, and collect all unique residues
    for design_id, val in descriptor_values.items():
        residues = {r.strip() for r in str(val).split(",") if r.strip()}
        design_to_residues[design_id] = residues
        unique_residues.update(residues)
    unique_residues = sorted(unique_residues)
    # Remove "None" if present
    if "None" in unique_residues:
        unique_residues.remove("None")
    bool_data = []

    # For each design, create a boolean row for presence of each residue
    for design_id in descriptor_values.index:
        row = [res in design_to_residues[design_id] for res in unique_residues]
        bool_data.append(row)
    bool_df = pd.DataFrame(bool_data, index=descriptor_values.index, columns=unique_residues)

    # Add column for number of residues present in each design and sort
    bool_df["# residues present"] = bool_df.sum(axis=1)
    bool_df = bool_df.sort_values(by="# residues present", ascending=False)

    caption = f"Designs (rows) vs residues (columns) for **{descriptor.name}**"
    column_config = {}
    column_config["# residues present"] = st.column_config.ProgressColumn(
        "# residues present",
        format="%d",
        min_value=0,
        max_value=len(unique_residues),
        width="small",
        help=f"Number of {descriptor.name} present in the design",
    )
    formatted_strings = {
        design_id: f"{design_id} | #residues={row['# residues present']}" for design_id, row in bool_df.iterrows()
    }
    format_func = lambda design_id: formatted_strings.get(design_id, str(design_id))
    return bool_df, column_config, caption, format_func



from ovo.core.database.models import NumericGlobalDescriptor, StructureFileDescriptor


# Hardcoded test options for refolding
REFOLDING_TESTS_SCAFFOLD = {
    "af2_model_1_ptm_nt_3rec": (
        "AF2 monomer, no template, 3 recycles",
        "AlphaFold2 model_1_ptm (monomer model), initial guess with no template input. Low risk of over-confidence.",
    ),
    "af2_model_1_ptm_ft_3rec": (
        "AF2 monomer, fixed motif template, 3 recycles",
        "AlphaFold2 model_1_ptm (monomer model), initial guess with template input for fixed regions. Medium risk of over-confidence.",
    ),
    "af2_model_1_multimer_nt_3rec": (
        "AF2 multimer, no template, 3 recycles",
        "AlphaFold2 model_1_multimer, initial guess with no template input. Low risk of over-confidence.",
    ),
    "af2_model_1_multimer_ft_3rec": (
        "AF2 multimer, fixed motif template, 3 recycles",
        "AlphaFold2 model_1_multimer, initial guess with template input for fixed regions. Medium risk of over-confidence.",
    ),
    "esmfold": ("ESMFold, 4 recycles", "Unbiased, sequence input only, low risk of over-confidence."),
}

REFOLDING_TESTS_BINDER = {
    "af2_model_1_ptm_tt_3rec": (
        "AF2 monomer, target template, 3 recycles",
        "AlphaFold2 model_1_ptm (monomer model) with residue index offset to predict two separate chains, initial guess with template input for target chain. Low risk of over-confidence.",
    ),
    "af2_model_1_ptm_tbt_3rec": (
        "AF2 monomer, target & binder template, 3 recycles",
        "AlphaFold2 model_1_ptm (monomer model) with residue index offset to predict two separate chains, initial guess with separate template input for target chain and binder chain. Medium risk of over-confidence.",
    ),
    "af2_model_1_ptm_ct_3rec": (
        "AF2 monomer, complex template, 3 recycles",
        "AlphaFold2 model_1_ptm (monomer model) with residue index offset to predict two separate chains, initial guess with template input for whole target and binder complex. High risk of over-confidence.",
    ),
    "af2_model_1_multimer_tt_3rec": (
        "AF2 multimer, target template, 3 recycles",
        "AlphaFold2 model_1_multimer, initial guess with template input for target chain. Low risk of over-confidence.",
    ),
    "af2_model_1_multimer_tbt_3rec": (
        "AF2 multimer, target & binder template, 3 recycles",
        "AlphaFold2 model_1_multimer, initial guess with separate template input for target chain and binder chain. Medium risk of over-confidence.",
    ),
    "af2_model_1_multimer_ct_3rec": (
        "AF2 multimer, complex template, 3 recycles",
        "AlphaFold2 model_1_multimer, initial guess with template input for whole target and binder complex. High risk of over-confidence.",
    ),
}


REFOLDING_TESTS_BY_TYPE = {
    "scaffold": REFOLDING_TESTS_SCAFFOLD,
    "binder": REFOLDING_TESTS_BINDER,
}

REFOLDING_TESTS = {k: v for tests in REFOLDING_TESTS_BY_TYPE.values() for k, v in tests.items()}

# ESMFold descriptors

ESMFOLD_STRUCTURE_PATH = StructureFileDescriptor(
    name="ESMFold prediction",
    description="Structure predicted by ESMFold, using only sequence input",
    tool="ESMFold",
    key="refolding|esmfold|esmfold_predicted_structure_path",
    structure_type="prediction",
    b_factor_value="fractional_plddt",
)

ESMFOLD_PLDDT = NumericGlobalDescriptor(
    name="ESMFold pLDDT",
    description="Average pLDDT confidence score of the whole structure (0 = worst, 100 = best)",
    tool="ESMFold",
    key="refolding|esmfold|pLDDT",
    min_value=0,
    max_value=100,
    comparison="higher_is_better",
    color_scale="plddt",
)

ESMFOLD_PTM = NumericGlobalDescriptor(
    name="ESMFold pTM score",
    description="Predicted TM score of the full structure (0 = worst, 1 = best)",
    tool="ESMFold",
    key="refolding|esmfold|pTM",
    min_value=0,
    max_value=1,
    comparison="higher_is_better",
)

ESMFOLD_PAE = NumericGlobalDescriptor(
    name="ESMFold PAE",
    description="Average predicted absolute error of the whole structure (in Angstrom)",
    unit="Å",
    tool="ESMFold",
    key="refolding|esmfold|pAE",
    min_value=0,
    comparison="lower_is_better",
    color_scale="pae",
)

ESMFOLD_DESIGN_BACKBONE_RMSD = NumericGlobalDescriptor(
    name="ESMFold Design Backbone RMSD",
    description="Aligned RMSD between the backbone of the designed structure and its ESMFold prediction",
    unit="Å",
    tool="ESMFold",
    key="refolding|esmfold|RMSD_backbone",
    min_value=0,
    comparison="lower_is_better",
    color_scale="rmsd",
)

ESMFOLD_DESIGN_ALL_ATOM_RMSD = NumericGlobalDescriptor(
    name="ESMFold Design All-atom RMSD",
    description="Aligned RMSD between all atoms of the designed structure and its ESMFold prediction",
    unit="Å",
    tool="ESMFold",
    key="refolding|esmfold|RMSD_all_atom",
    min_value=0,
    comparison="lower_is_better",
    color_scale="rmsd",
)

ESMFOLD_DESCRIPTORS = [
    ESMFOLD_DESIGN_BACKBONE_RMSD,
    ESMFOLD_PAE,
    ESMFOLD_PLDDT,
    ESMFOLD_PTM,
    ESMFOLD_DESIGN_ALL_ATOM_RMSD,
]


# INITIAL GUESS DESCRIPTORS
AF2_PRIMARY_STRUCTURE_PATH = StructureFileDescriptor(
    name="AlphaFold2 Initial Guess prediction",
    description="AlphaFold2 structure prediction using Initial Guess protocol initialized with designed structure, with optional structure template input",
    tool="AF2 Initial Guess",
    key="refolding|af2_primary|af2_structure_path",
    structure_type="prediction",
    b_factor_value="plddt",
)

AF2_PRIMARY_TARGET_ALIGNED_BINDER_RMSD = NumericGlobalDescriptor(
    name="AF2 Target-aligned Binder RMSD",
    description="Target-aligned binder RMSD between Ca backbone atoms of design and AF2 prediction",
    unit="Å",
    tool="AF2 Initial Guess",
    key="refolding|af2_primary|target_aligned_binder_rmsd",
    min_value=0,
    comparison="lower_is_better",
    color_scale="rmsd",
)

AF2_PRIMARY_IPAE = NumericGlobalDescriptor(
    name="AF2 iPAE",
    description="AlphaFold2 interaction PAE, predicted aligned error of quadrants of the PAE matrix corresponding to all pairs of residues between the interacting chains (in Angstrom)",
    unit="Å",
    tool="AF2 Initial Guess",
    key="refolding|af2_primary|ipae",
    min_value=0,
    comparison="lower_is_better",
    color_scale="pae",
)

AF2_PRIMARY_IPTM = NumericGlobalDescriptor(
    name="AF2 ipTM score",
    description="AlphaFold2 interface predicted TM score (0 = worst, 1 = best) based on all pairs of residues between the interacting chains",
    tool="AF2 Initial Guess",
    key="refolding|af2_primary|iptm",
    min_value=0,
    max_value=1,
    comparison="higher_is_better",
)

AF2_PRIMARY_BINDER_PAE = NumericGlobalDescriptor(
    name="AF2 Binder PAE",
    description="Predicted aligned error of the binder chain (in Angstrom)",
    unit="Å",
    tool="AF2 Initial Guess",
    key="refolding|af2_primary|binder_pae",
    min_value=0,
    comparison="lower_is_better",
    color_scale="pae",
)

AF2_PRIMARY_PLDDT_BINDER = NumericGlobalDescriptor(
    name="AF2 Binder pLDDT",
    description="Average pLDDT confidence score of the binder chain (0 = worst, 100 = best)",
    tool="AF2 Initial Guess",
    key="refolding|af2_primary|binder_plddt",
    min_value=0,
    max_value=100,
    comparison="higher_is_better",
    color_scale="plddt",
)

AF2_PRIMARY_DESIGN_RMSD = NumericGlobalDescriptor(
    name="AF2 Design RMSD",
    description="Aligned RMSD between the backbone of the designed structure and its AF2 prediction",
    unit="Å",
    tool="AF2 Initial Guess",
    key="refolding|af2_primary|design_backbone_rmsd",
    min_value=0,
    comparison="lower_is_better",
    color_scale="rmsd",
)

AF2_PRIMARY_NATIVE_MOTIF_RMSD = NumericGlobalDescriptor(
    name="AF2 Native Motif RMSD",
    description="Aligned RMSD between all atoms of the fixed input motif and its AF2 prediction",
    unit="Å",
    tool="AF2 Initial Guess",
    key="refolding|af2_primary|native_motif_rmsd",
    min_value=0,
    comparison="lower_is_better",
    color_scale="rmsd",
)

AF2_PRIMARY_PTM = NumericGlobalDescriptor(
    name="AF2 pTM score",
    description="Predicted TM score of the full structure (0 = worst, 1 = best)",
    tool="AF2 Initial Guess",
    key="refolding|af2_primary|ptm",
    min_value=0,
    max_value=1,
    comparison="higher_is_better",
)


AF2_PRIMARY_PAE = NumericGlobalDescriptor(
    name="AF2 PAE",
    description="Average predicted absolute error of the whole structure (in Angstrom)",
    unit="Å",
    tool="AF2 Initial Guess",
    key="refolding|af2_primary|pae",
    min_value=0,
    comparison="lower_is_better",
    color_scale="pae",
)

AF2_PRIMARY_PLDDT = NumericGlobalDescriptor(
    name="AF2 pLDDT",
    description="Average pLDDT score of the whole structure (0 = worst, 100 = best)",
    tool="AF2 Initial Guess",
    key="refolding|af2_primary|plddt",
    min_value=0,
    max_value=100,
    comparison="higher_is_better",
    color_scale="plddt",
)


# Initial guess descriptors
AF2_PRIMARY_DESCRIPTORS = [
    # shared
    AF2_PRIMARY_IPAE,
    AF2_PRIMARY_IPTM,
    AF2_PRIMARY_PTM,
    # binder
    AF2_PRIMARY_BINDER_PAE,
    AF2_PRIMARY_PLDDT_BINDER,
    AF2_PRIMARY_TARGET_ALIGNED_BINDER_RMSD,
    # scaffold
    AF2_PRIMARY_DESIGN_RMSD,
    AF2_PRIMARY_NATIVE_MOTIF_RMSD,
    AF2_PRIMARY_PAE,
    AF2_PRIMARY_PLDDT,
]
AF2_STRUCTURE_PATHS = [AF2_PRIMARY_STRUCTURE_PATH]


# Refolding
REFOLDING_DESCRIPTORS = [*ESMFOLD_DESCRIPTORS, *AF2_PRIMARY_DESCRIPTORS, ESMFOLD_STRUCTURE_PATH, *AF2_STRUCTURE_PATHS]


for test, (label, description) in REFOLDING_TESTS_SCAFFOLD.items():
    REFOLDING_DESCRIPTORS += [
        NumericGlobalDescriptor(
            name="AF2 Design RMSD",
            description=f"Aligned RMSD between the backbone of the designed structure and its AF2 prediction using {description}",
            unit="Å",
            tool=f"AlphaFold2 ({label})",
            key=f"refolding|{test}|design_backbone_rmsd",
            min_value=0,
            comparison="lower_is_better",
            color_scale="rmsd",
        ),
        NumericGlobalDescriptor(
            name="AF2 Native Motif RMSD",
            description=f"Aligned RMSD between all atoms of the fixed input motif and its AF2 prediction using {description}",
            unit="Å",
            tool=f"AlphaFold2 ({label})",
            key=f"refolding|{test}|native_motif_rmsd",
            min_value=0,
            comparison="lower_is_better",
            color_scale="rmsd",
        ),
        NumericGlobalDescriptor(
            name=f"AF2 iPAE",
            description=f"AlphaFold2 interaction PAE, predicted aligned error of quadrants of the PAE matrix corresponding to all pairs of residues between the interacting chains (in Angstrom) using {description}",
            unit="Å",
            tool=f"AlphaFold2 ({label})",
            key=f"refolding|{test}|ipae",
            min_value=0,
            comparison="lower_is_better",
            color_scale="pae",
        ),
        NumericGlobalDescriptor(
            name=f"AF2 pLDDT",
            description=f"Average pLDDT score of the whole structure (0 = worst, 100 = best) using {description}",
            tool=f"AlphaFold2 ({label})",
            key=f"refolding|{test}|plddt",
            min_value=0,
            max_value=100,
            comparison="higher_is_better",
            color_scale="plddt",
        ),
        NumericGlobalDescriptor(
            name=f"AF2 PAE",
            description=f"Average predicted absolute error of the whole structure (in Angstrom) using {description}",
            unit="Å",
            tool=f"AlphaFold2 ({label})",
            key=f"refolding|{test}|pae",
            min_value=0,
            comparison="lower_is_better",
            color_scale="pae",
        ),
        NumericGlobalDescriptor(
            name=f"AF2 ipTM score",
            description=f"AlphaFold2 interface predicted TM score (0 = worst, 1 = best) based on all pairs of residues between the interacting chains using {description}",
            tool=f"AlphaFold2 ({label})",
            key=f"refolding|{test}|iptm",
            min_value=0,
            max_value=1,
            comparison="higher_is_better",
        ),
        NumericGlobalDescriptor(
            name="AF2 pTM score",
            description=f"Predicted TM score of the full structure (0 = worst, 1 = best) using {description}",
            tool=f"AlphaFold2 ({label})",
            key=f"refolding|{test}|ptm",
            min_value=0,
            max_value=1,
            comparison="higher_is_better",
        ),
    ]


for test, (label, description) in REFOLDING_TESTS_BINDER.items():
    REFOLDING_DESCRIPTORS += [
        NumericGlobalDescriptor(
            name="AF2 Target-aligned Binder RMSD",
            description=f"Target-aligned binder RMSD between Ca backbone atoms of design and AF2 prediction using {description}",
            unit="Å",
            tool=f"AlphaFold2 ({label})",
            key=f"refolding|{test}|target_aligned_binder_rmsd",
            min_value=0,
            comparison="lower_is_better",
            color_scale="rmsd",
        ),
        NumericGlobalDescriptor(
            name=f"AF2 iPAE",
            description=f"AlphaFold2 interaction PAE, predicted aligned error of quadrants of the PAE matrix corresponding to all pairs of residues between the interacting chains (in Angstrom) using {description}",
            unit="Å",
            tool=f"AlphaFold2 ({label})",
            key=f"refolding|{test}|ipae",
            min_value=0,
            comparison="lower_is_better",
            color_scale="pae",
        ),
        NumericGlobalDescriptor(
            name="AF2 Binder pLDDT",
            description=f"Average pLDDT confidence score of the binder chain (0 = worst, 100 = best) using {description}",
            tool=f"AlphaFold2 ({label})",
            key=f"refolding|{test}|binder_plddt",
            min_value=0,
            max_value=100,
            comparison="higher_is_better",
            color_scale="plddt",
        ),
        NumericGlobalDescriptor(
            name="AF2 Binder PAE",
            description=f"Predicted aligned error of the binder chain (in Angstrom) using {description}",
            unit="Å",
            tool=f"AlphaFold2 ({label})",
            key=f"refolding|{test}|binder_pae",
            min_value=0,
            comparison="lower_is_better",
            color_scale="pae",
        ),
        NumericGlobalDescriptor(
            name=f"AF2 ipTM score",
            description=f"AlphaFold2 interface predicted TM score (0 = worst, 1 = best) based on all pairs of residues between the interacting chains using {description}",
            tool=f"AlphaFold2 ({label})",
            key=f"refolding|{test}|iptm",
            min_value=0,
            max_value=1,
            comparison="higher_is_better",
        ),
        NumericGlobalDescriptor(
            name="AF2 pTM score",
            description=f"Predicted TM score of the full structure (0 = worst, 1 = best) using {description}",
            tool=f"AlphaFold2 ({label})",
            key=f"refolding|{test}|ptm",
            min_value=0,
            max_value=1,
            comparison="higher_is_better",
        ),
    ]

DESCRIPTORS = REFOLDING_DESCRIPTORS

PRESETS = [
    {
        "label": "AF2 PAE & RMSD",
        "x": AF2_PRIMARY_PAE,
        "y": AF2_PRIMARY_DESIGN_RMSD,
    },
    {
        "label": "AF2 iPAE & RMSD",
        "x": AF2_PRIMARY_IPAE,
        "y": AF2_PRIMARY_TARGET_ALIGNED_BINDER_RMSD,
    },
    {
        "label": "ESMFold PAE & RMSD",
        "x": ESMFOLD_PAE,
        "y": ESMFOLD_DESIGN_BACKBONE_RMSD,
    },
]



from sqlalchemy.exc import NoResultFound

from ovo import db, get_username
from ovo.core.database.models import Project, Round


PERSONAL_PROJECT_NAME = "Personal project: {author}"


def get_or_create_personal_project() -> Project:
    author = get_username()
    return get_or_create_project(
        PERSONAL_PROJECT_NAME.format(author=author),
        public=False,
        filters=dict(public=False, author=author),
    )


def get_or_create_project(project_name: str, public=False, filters=None) -> Project:
    """Get or create a project by name."""
    author = get_username()
    try:
        return db.get(Project, name=project_name, **(filters or {}))
    except NoResultFound:
        project = Project(name=project_name, public=public, author=author)
        db.save(project)
        return project


def get_or_create_project_round(project_name: str, round_name: str) -> tuple[Project, Round]:
    project = get_or_create_project(project_name)
    try:
        project_round = db.get(Round, name=round_name, project_id=project.id)
    except NoResultFound:
        project_round = Round(name=round_name, project_id=project.id, author=get_username())
        db.save(project_round)
    return project, project_round



import os

import streamlit as st
from typing import Dict, List, Callable
import json
import pandas as pd
import numpy as np
import altair as alt

from ovo import Descriptor
from ovo.core.logic.proteinqc_logic import get_descriptor_plot_setting, get_cmap
from ovo.core.utils.colors import darken

FLAG_ICONS = {
    "missing": ":material/question_mark:",
    "green": ":material/check_circle:",
    "yellow": ":material/warning:",
    "orange": ":material/error:",
}

SOURCE_OPTIONS = {
    "all": "PDB: All structures",
    "non-human_only": "PDB: Non-human only",
    "human_only": "PDB: Human only",
    "human+non-human": "PDB: Human or human-interacting",
    "short (<=100)": "PDB: Short (<= 100aa)",
    "medium (101-300)": "PDB: Medium length (101-300aa)",
    "long (301-1000)": "PDB: Long (301-1000aa)",
    "very long (>1000)": "PDB: Very long (> 1000aa)",
}


@st.cache_data()
def load_histograms() -> Dict[str, Dict[str, Dict[str, List[float]]]]:
    """
    Load precomputed PDB histograms for all tools.
    The histograms are computed for the data between 1st and 99th percentile.

    Returns:
    {'descriptor': {'all': {'x': [0, 1, 2, ...], 'y': [0.1, 0.2, ...]}, 'non-human_only': ..., 'human_only': ..., 'human+non-human}, ...}}}}
    x - bin edges
    y - bin heights
    """
    histograms = {}
    for toolname in [
        "esm_1v",
        "esm_if",
        "proteinsol",
        "sequence_composition",
        "protparams",
        "dssp",
        "length",
        "mdanalysis",
        "peppatch",
    ]:
        data = load_plots_data(f"histograms/pdb/{toolname}_q01_q99.json")
        histograms.update({f"proteinqc|{key}": value for key, value in data.items()})
    return histograms


def load_plots_data(file_path: str) -> dict:
    ovo_resources_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "resources"))
    with open(os.path.join(ovo_resources_path, file_path), "r") as f:
        data = json.load(f)
    return data


@st.cache_data
def get_histogram_alt(
    bins: np.ndarray,
    y: np.ndarray,
    color_by: str,
    thresholds: List[float] = None,
    reverse_colors: bool = False,
    width: int = None,
    height: int = None,
    title: str = None,
    x_label: str = None,
    y_label: str = None,
    value: float = None,
    value_label: str = None,
) -> alt.Chart:
    """
    Create Altair histogram from precomputed data.

    Args:
    bins: bin edges in the precomputed histograms
    y: bin heights in the precomputed histograms
    color_by: coloring method, one of 'reference', 'thresholds', 'plddt', 'rmsd', or other color schemes supported by get_cmap
    thresholds: list of thresholds for color_by='thresholds'
    reverse_colors: reverse the color scale
    width: width of the chart
    height: height of the chart
    title: title of the chart
    x_label: label for the x-axis
    y_label: label for the y-axis
    value: single value to highlight on the histogram
    value_label: label for the highlighted value
    """

    # Create DataFrame for Altair
    bin_centers = bins[:-1] + np.diff(bins) / 2
    total_height_sum = np.sum(y)  # Sum of heights for normalization
    data = pd.DataFrame(
        {
            "bin_center": bin_centers,
            "height": y,
            "normalized_height": y / total_height_sum,  # Normalized for color mapping
        }
    )
    max_height = data["normalized_height"].max()
    darken_factor = 1.2
    cmap: None | Callable = None
    if color_by == "reference":
        color = alt.value("#bbbbbb")
    elif (color_by == "thresholds" and thresholds is not None) or color_by in ["plddt", "rmsd"]:
        # use colormap if applicable in histogram
        min_val = data["bin_center"].min()
        max_val = data["bin_center"].max()
        cmap = get_cmap(color_by, min_val, max_val, thresholds, reverse_colors=reverse_colors)
        data["flag_color"] = data["bin_center"].apply(cmap).apply(darken, factor=darken_factor)
        color = alt.Color("flag_color:N", legend=None, scale=None)
    else:
        color = alt.value("#444444")

    bar_width = min(5.0, 1 + 150 / len(bins))
    chart = (
        alt.Chart(data)
        .mark_bar(width=bar_width)
        .encode(
            x=alt.X("bin_center:Q", title=None, axis=alt.Axis(title=x_label)),
            y=alt.Y(
                "normalized_height:Q", axis=alt.Axis(title=y_label, labels=False, ticks=False, domain=False)
            ),  # Hide y axis label and ticks
            color=color,
            tooltip=[],
        )
    )

    if cmap:
        band_data = pd.DataFrame(
            {
                "x0": bins[:-1],
                "x1": bins[1:],
                "y0": max_height * -0.05,
                "y1": max_height * -0.09,
            }
        )
        band_data["color"] = ((band_data["x0"] + band_data["x1"]) / 2).apply(cmap).apply(darken, factor=darken_factor)
        gradient_band = (
            alt.Chart(band_data)
            .mark_rect(opacity=0.5)
            .encode(
                x="x0:Q",
                x2="x1:Q",
                y="y0:Q",
                y2="y1:Q",
                color=alt.Color("color:N", scale=None, legend=None),
                tooltip=[],
            )
        )

        chart = (chart + gradient_band).resolve_scale(color="independent")

    if value is not None:
        value_data = pd.DataFrame({"x": [value], "y": [max_height * -0.15], "label": value_label})
        triangle = (
            alt.Chart(value_data)
            .mark_point(shape="triangle-up", color="black", size=90, filled=True)
            .encode(x="x:Q", y="y:Q", tooltip=[])
        )

        chart = chart + triangle

        if value_label:
            align = "right" if value >= np.mean(bins) else "left"
            label = (
                alt.Chart(value_data)
                .mark_text(align=align, baseline="top", fontSize=12, dx=8 * (-1 if align == "right" else 1))
                .encode(
                    x="x:Q",
                    y="y:Q",
                    text="label:N",
                    tooltip=[],
                )
            )
            chart = chart + label

    if title:
        chart = chart.properties(title=title)

    if width:
        chart = chart.properties(width=width)

    if height:
        chart = chart.properties(height=height)

    return chart


def update_bin_edges(values: List[float], bins: List[float], y: list[float]) -> tuple[np.ndarray, np.ndarray]:
    """
    Given bins for a histogram and a list of values, extend the bins if the values are outside the range of the bins.
    """

    bin_width = np.diff(bins[:2])[0]

    min_bin = bins[0]
    max_bin = bins[-1]

    # If the second distribution has a wider range, extend the bins
    if (min_value := np.min(values)) < min_bin:
        # Create extra bins on the left
        extra_left = np.arange(min_bin - bin_width, min_value - bin_width, -bin_width)[::-1]
        bins = np.concatenate([extra_left, bins])
        y = np.concatenate([np.zeros(len(extra_left)), y])

    if (max_value := np.max(values)) > max_bin:
        # Create extra bins on the right
        extra_right = np.arange(max_bin + bin_width, max_value + bin_width, bin_width)
        bins = np.concatenate([bins, extra_right])
        y = np.concatenate([y, np.zeros(len(extra_right))])

    return bins, y


def descriptor_histograms(
    descriptor: Descriptor,
    histogram_source: str,
    values: List[float],
    value: float = None,
    value_label: str = None,
    color_by: str = "default",
    precomputed_histogram_height: int = 150,
    pool_histogram_height: int = 150,
    width: int = 600,
):
    """
    Create a combined histogram plot for a descriptor. The top histogram is the precomputed histogram and the bottom histogram is the pool histogram from the passed values.

    Parameters:
    descriptor (Descriptor): descriptor object
    histogram_source (str): Source of the histogram
    values (List[float]): Values for the pool/selected designs histogram
    value (float): Single value to highlight on the histogram
    value_label (str): Label for the highlighted value
    color_by (str): Color by 'thresholds', None, or values returned by get_descriptor_plot_setting
    precomputed_histogram_height (int): Height of the precomputed histogram
    pool_histogram_height (int): Height of the pool histogram
    width (int): Maximum width of the chart in pixels
    """
    thresholds, reverse_colors, default_color_by = get_descriptor_plot_setting(descriptor)
    if color_by == "default":
        color_by = default_color_by

    histograms = load_histograms()
    if descriptor.key not in histograms:
        bins = np.linspace(np.min(values), np.max(values), 50)
        fig1 = None
    else:
        data = histograms[descriptor.key][histogram_source]
        bins, ref_y = update_bin_edges(values, data["x"], data["y"])

        # Precomputed histogram
        fig1 = get_histogram_alt(
            bins=bins,
            y=ref_y,
            y_label="Reference",
            height=precomputed_histogram_height,
            width=width,
            color_by="reference",
        )

    fig2 = get_histogram_alt(
        bins=bins,
        y=np.histogram(values, bins)[0],
        x_label=descriptor.name,
        y_label="Designs",
        height=pool_histogram_height,
        width=width,
        color_by=color_by,
        thresholds=thresholds,
        reverse_colors=reverse_colors,
        value=value,
        value_label=value_label,
    )

    if not fig1:
        # Return only the pool histogram
        return fig2.properties(height=pool_histogram_height + 80).configure_axis(grid=False)
    # Stack the two histograms

    return (
        alt.vconcat(fig1, fig2)
        .resolve_legend(color="independent")
        .resolve_scale(
            x="shared",  # Share the x-axis scale between the two histograms
            y="independent",  # Independent y-axis to scale correctly
        )
        .configure_axis(grid=False)
    )


@st.cache_data
def format_threshold_value(descriptor_name: str, value) -> str | None:
    """
    Format the threshold values for display.

    If the descriptor is a percentage (0-1), multiply by 100 and add a percentage sign.
    If the descriptor is 'Sequence length', round to the nearest integer.
    Otherwise, round to two decimal places.
    """
    if pd.isna(value) or value == "":
        return None
    # TODO store this logic in the Descriptor object instead
    if "%" in descriptor_name:
        return f"{value:.2f} %"
    elif descriptor_name == "Sequence length":
        return f"{value:.0f}"
    else:
        return f"{value:.2f}"


def source_selectbox(value: str, key_prefix: str):
    options = list(SOURCE_OPTIONS)
    return st.selectbox(
        label="Reference distribution",
        format_func=SOURCE_OPTIONS.get,
        help="""
        Select the background distribution to compare the pool/selection distribution to.  
        The **"all"** option includes all PDB structures.  
        The **"non-human only"** option includes only non-human proteins.  
        The **"human only"** option includes only human proteins.  
        The **"human+non-human"** option includes proteins from human+non-human complexes.  
        The **"short"** (≤100), **"medium"** (101-300), **"long"** (301-1000), and **"very long"** (>1000) options include proteins with the corresponding sequence length.""",
        index=options.index(value) if value else 0,
        key=f"{key_prefix}_source",
        options=options,
    )
